/////////////////////////////////////////////////////////////////////////////
///
/// @file quaditer.cpp
///
/// This file contains the main program which processes the command-line
/// arguments and runs the computation for the 2020 paper.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2006-2021 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <http://www.gnu.org/licenses/>.

// Started on June 11, 2013. Last revision: March 3, 2021.

// NOTE: This program uses the MPFR library (and a few files from CHomP).


// standard library headers
#include <string>
#include <sstream>
#include <vector>
#include <queue>
#include <algorithm>
#include <exception>

// headers for multi-precision rigorous arithmetic
#include <gmp.h>
#include <mpfr.h>
#include "mpfrsupp.h"

// selected files from the CHomP library
#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/system/timeused.h"
#include "chomp/system/arg.h"


// --------------------------------------------------
// -------------------- OVERTURE --------------------
// --------------------------------------------------

/// The title of the main program.
const char *title = "\
QuadraticMapIteration 1.01, Mrch 03, 2021. (C) 2006-2021 by Pawel Pilarczyk.\n\
This is free software. No warranty. Consult 'license.txt' for details.";

/// The help information about the main program.
const char *helpinfo = "\
This is a computer program for rigorous computation of iterates of the\n\
quadratic map, as described in the paper by A. Golmakani, C.E. Koudjinan,\n\
S. Luzzatto, and P. Pilarczyk.\n\
Command line arguments (using default values if the argument is omitted):\n\
-l N - the left endpoint of the parameter interval (obligatory argument),\n\
-r N - the right endpoint of the parameter interval (obligatory argument),\n\
-u N - start with a uniform subdivision into N intervals (1),\n\
-d N - the radius of the critical neighborhood, a.k.a. delta (0.001),\n\
-p N - the binary precision of high-prec computations (1000, max 100000),\n\
-b N - the number of bisection steps when excluding (30),\n\
-n N - the maximal allowed number of iterations (1000, max 100000),\n\
-f N - the minimal interval fraction required to continue iterating (0),\n\
-w N - the minimal width for an interval, relative to the input (10^-10),\n\
-c N - the number of contractions encountered to stop iterating (0=off),\n\
-i N - the maximal number of intervals to process (0 = no limit),\n\
-q N - the maximal allowed queue size (0 = no limit),\n\
-m N - the maximal number of messages to display (0 = no limit),\n\
-t N - the maximal running time in seconds (10; use 0 for no limit),\n\
-N N - the minimal number of iterates to reach by all to stop (0=unused),\n\
-rN N - the required number of iterates to count as an 'escape',\n\
-rW N - the required image width to count as an 'escape',\n\
-o N - the number of decimal digits when writing numbers (8, max 1000),\n\
-v N - the verbosity of program's output (default 0, max 100),\n\
--help - show this brief help information and exit.\n\
Please, see the paper, the Web page, and the source code for details.\n\
For more information ask the author at http://www.pawelpilarczyk.com/.";

// Features considered for the next versions:
// - Verbosity as flags: turning on/off individual types of messages.


// --------------------------------------------------
// ------------------- verbosity --------------------
// --------------------------------------------------

/// Minimal verbosity levels requied to show additional types of information.
namespace Verbosity
{

// messages that are shown when certain verbosity level is requested
int basicSummary (1);
int measureSummary (3);
int computationSummary (5);
int queueSummary (10);
int programArguments (13);
int stopReason (15);
int periodicSummary (17);
int init (20);
int addToQueue (22);
int signMismatch (25);
int omegaGood (27);
int omegaHitDelta (30);
int omegaSplit (32);
int innerBoundEmpty (35);
int exclusions (50);
int abandonedAttracting (55);
int abandonedLong (60);
int abandonedSmall (65);
int leftInQueue (70);
int iterates (80);
int boundsApart (90);
int contractionHeuristics (95);
int allComputedIntervals (99);

// data that is only shown upon special request
int exclusionData (1000);
int hitData (1000);
int goodData (1000);
int iteratedData (1000);
int derivativeData (1000);

} /* namespace Verbosity */


/// A global switch on whether the intervals that come from halving
/// problematic intervals should be processed first. If so then the queue
/// is cleared of "bad" intervals as soon as possible, even at the cost
/// of not processing "good" intervals for some time.
/// Note that the minimal width of the intervals or the maximal number
/// of iterates allowed must be set to some reasonable values,
/// or otherwise this switch could make the program loop almost indefinitely,
/// halving some intervals into tiny, tiny pieces.
bool problemsFirst = false;


// --------------------------------------------------
// -------------------- REASONS ---------------------
// --------------------------------------------------

/// Reasons for adding an interval to the queue.
namespace Reason
{
	/// reason: new interval added at initialization
	int init (0);

	/// reason: after chopping due to collision of omega_n with Delta
	int chop (1);

	/// reason: sign mismatch with C
	int problemC (2);

	/// reason: sign mismatch with F
	int problemF (3);

} /* namespace Reason */

/// A class for counting the reasons.
class ReasonCounter
{
public:
	/// Increases the counter of the given reason.
	/// Returns the new value of the counter.
	int inc (int reason);

	/// Decreases the counter of the given reason.
	/// Returns the new value of the counter.
	int dec (int reason);

	/// Returns the counter value for the given reason.
	int get (int reason) const;

	/// Returns the number of reasons accounted for.
	size_t size () const;

private:
	/// The internal array of counters of the reasons.
	std::vector<int> _counters;

}; /* class ReasonCounter */

// --------------------------------------------------

inline int ReasonCounter::inc (int reason)
{
	if ((reason < 0) || (reason > 1000))
		throw "Wrong reason when increasing the counter.";
	while (static_cast<int> (_counters. size ()) <= reason)
		_counters. push_back (0);
	++ _counters [reason];
	return _counters [reason];
} /* ReasonCounter::inc */

inline int ReasonCounter::dec (int reason)
{
	if ((reason < 0) || (reason >= static_cast<int> (_counters. size ())))
		throw "Wrong reason when decreasing the counter.";
	if (_counters [reason] <= 0)
		throw "Trying to decrease reason counter that is zero.";
	-- _counters [reason];
	return _counters [reason];
} /* ReasonCounter::dec */

inline int ReasonCounter::get (int reason) const
{
	if ((reason < 0) || (reason >= static_cast<int> (_counters. size ())))
		return 0;
	return _counters [reason];
} /* ReasonCounter::get */

inline size_t ReasonCounter::size () const
{
	return _counters. size ();
} /* ReasonCounter::size */


// --------------------------------------------------
// ----------------- INTERVAL INFO ------------------
// --------------------------------------------------

/// A class for storing two pointers to interval endpoints,
/// as well as the number of previous iterates,
/// some data (e.g. the reason for adding it to the queue),
/// and the consecutive number of the item.
/// Important information: The points themselves are NOT
/// stored in this record but POINTERS to the MPFR numbers.
/// Therefore, any procedure that uses this structure
/// must take care of allocating and deallocating memory.
class IntervalInfo
{
public:
	/// The default constructor.
	/// Note: The default destructor, copy constructor,
	/// and assignment operator are fine.
	IntervalInfo (int N, int C, int D,
		auto_mpfr *left, auto_mpfr *right, double width);

	/// Returns the number N that was previously defined.
	int getN () const;

	/// Returns the number C that was previously defined.
	int getC () const;

	/// Returns the number D that was previously defined.
	int getD () const;

	/// Returns the width of the interval, rounded for faster processing.
	double getW () const;

	/// Returns a pointer to the left bound.
	auto_mpfr *getLeft () const;

	/// Returns a pointer to the left bound.
	auto_mpfr *getRight () const;

	/// The operator for comparing the intervals based on their priority.
	bool operator < (const IntervalInfo &other) const;

private:
	/// The number of times the interval was previously iterated.
	int _N;

	/// The consecutive number of the item that indicates the order
	/// in which the items were added to the list.
	int _C;

	/// The data that is stored together with the interval.
	int _D;

	/// The width of the interval rounded to a double-precision
	/// floating point number.
	double _W;

	/// A pointer to the left bound of the interval.
	auto_mpfr *_left;

	/// A pointer to the right bound of the interval.
	auto_mpfr *_right;

}; /* class IntervalInfo */

// --------------------------------------------------

inline IntervalInfo::IntervalInfo (int N, int C, int D,
	auto_mpfr *left, auto_mpfr *right, double width):
	_N (N), _C (C), _D (D), _left (left), _right (right)
{
	if (width != 0)
		_W = width;
	else
	{
		auto_mpfr w;
		mpfr_sub (w, *right, *left, GMP_RNDN);
		_W = mpfr_get_d (w, GMP_RNDN);
	}
	return;
} /* IntervalInfo::IntervalInfo */

inline int IntervalInfo::getN () const
{
	return _N;
} /* IntervalInfo::getN */

inline int IntervalInfo::getC () const
{
	return _C;
} /* IntervalInfo::getC */

inline int IntervalInfo::getD () const
{
	return _D;
} /* IntervalInfo::getD */

inline double IntervalInfo::getW () const
{
	return _W;
} /* IntervalInfo::getW */

inline auto_mpfr *IntervalInfo::getLeft () const
{
	return _left;
} /* IntervalInfo::getLeft */

inline auto_mpfr *IntervalInfo::getRight () const
{
	return _right;
} /* IntervalInfo::getRight */

inline bool IntervalInfo::operator <
	(const IntervalInfo &other) const
{
	if (problemsFirst && ((this -> _D > 1) || (other. _D > 1)))
	{
		// take the intervals with highest reason codes first
		if (this -> _D != other. _D)
			return (this -> _D < other. _D);
		// take the intervals with smallest N first
		if (this -> _N != other. _N)
			return this -> _N > other. _N;
		// otherwise take the intervals in the order
		// in which they were enqueued
		return (this -> _C > other. _C);
	}
	else
	{
		// take the intervals with smallest N first
		if (this -> _N != other. _N)
			return this -> _N > other. _N;
		// take the widest intervals first
		if (this -> _W != other. _W)
			return this -> _W < other. _W;
		// take the intervals with highest reason codes first
		if (this -> _D != other. _D)
			return (this -> _D < other. _D);
		// otherwise take the intervals in the order
		// in which they were enqueued
		return (this -> _C > other. _C);
	}
} /* IntervalInfo::operator < */


// --------------------------------------------------
// ---------------- SET OF INTERVALS ----------------
// --------------------------------------------------

/// A set of intervals with the feature of joining adjacent intervals,
/// and with the feature of sorting the intervals.
/// The typical usage would be to create the set, add several intervals,
/// then sort the set and condense it, and finally read the intervals out.
/// One may also want to sort the set by size before reading the intervals.
class IntervalSet
{
public:
	/// The constructor of an empty set of intervals.
	IntervalSet ();

	/// The destructor.
	~IntervalSet ();

	/// Adds an interval to the collection.
	/// Appends the information about the interval
	/// but doesn't sort or join the intervals yet.
	void add (const auto_mpfr &leftOmega,
		const auto_mpfr &rightOmega, double width);

	/// Adds an interval to the collection.
	/// The memory for the endpoints was already allocated.
	void add (auto_mpfr *leftNew, auto_mpfr *rightNew, double width);

	/// Returns the number of intervals waiting in the queue.
	size_t size () const;

	/// Returns an interval at the given position in the set.
	/// Does not change the set of intervals.
	void get (size_t n, auto_mpfr *&leftOmega,
		auto_mpfr *&rightOmega, double *width, int *N) const;

	/// Sorts the intervals in the ascending order by their location.
	void sort_by_location ();

	/// Sorts the intervals in the descending order by their width.
	/// The longest interval comes first after this sorting.
	void sort_by_width ();

	/// Joins adjacent intervals in order to condense the reprsentation.
	/// The set of intervals must be sorted before this operation.
	void condense ();

private:
	/// The copy constructor is not allowed.
	IntervalSet (const IntervalSet &src);

	/// The assignment operator is not allowed.
	IntervalSet &operator = (const IntervalSet &src);

	/// The pairs of endpoints of the intervals
	/// and their approximate widths.
	/// The first integer, called N, represents the number of intervals
	/// that were joined together into this particular interval.
	/// Note that the other integers stored in the structure
	/// "IntervalInfo" are ignored here.
	std::vector<IntervalInfo> _intervals;

}; /* class IntervalSet */

// --------------------------------------------------

inline IntervalSet::IntervalSet ()
{
	return;
} /* IntervalSet::IntervalSet */

inline IntervalSet::~IntervalSet ()
{
	for (std::vector<IntervalInfo>::iterator it (_intervals. begin());
		it != _intervals.end(); ++ it)
	{
		IntervalInfo &ii = *it;
		delete ii. getLeft ();
		delete ii. getRight ();
	}
	return;
} /* IntervalSet::~IntervalSet */

inline IntervalSet::IntervalSet (const IntervalSet &)
{
	throw "The copy constructor not allowed in the IntervalSet class.";
	return;
} /* IntervalSet::IntervalSet */

inline IntervalSet &IntervalSet::operator = (const IntervalSet &)
{
	throw "The assignment operator not allowed in the IntervalSet class.";
	return *this;
} /* IntervalSet::operator = */

inline void IntervalSet::add (const auto_mpfr &leftOmega,
	const auto_mpfr &rightOmega, double width)
{
	// make a copy of the endpoints of the interval
	auto_mpfr *leftNew = new auto_mpfr;
	mpfr_set (*leftNew, leftOmega, GMP_RNDD);
	auto_mpfr *rightNew = new auto_mpfr;
	mpfr_set (*rightNew, rightOmega, GMP_RNDU);

	// add the record to the list
	_intervals. push_back (IntervalInfo (1, 0, 0,
		leftNew, rightNew, width));

	return;
} /* IntervalSet::add */

inline void IntervalSet::add (auto_mpfr *leftNew, auto_mpfr *rightNew,
	double width)
{
	// add the record to the list
	_intervals. push_back (IntervalInfo (1, 0, 0,
		leftNew, rightNew, width));

	return;
} /* IntervalSet::add */

inline size_t IntervalSet::size () const
{
	return _intervals. size ();
} /* IntervalSet::size */

inline void IntervalSet::get (size_t n, auto_mpfr *&leftOmega,
	auto_mpfr *&rightOmega, double *width, int *N) const
{
	// set the endpoints with the requested values
	if (n >= _intervals. size ())
		throw "Too high an index of an interval requested.";
	leftOmega = _intervals [n]. getLeft ();
	rightOmega = _intervals [n]. getRight ();
	//mpfr_set (leftOmega, *(_intervals [n]. getLeft ()), GMP_RNDD);
	//mpfr_set (rightOmega, *(_intervals [n]. getRight ()), GMP_RNDU);
	if (width)
		*width = _intervals [n]. getW ();
	if (N)
		*N = _intervals [n]. getN ();
	return;
} /* IntervalSet::get */

/// A helper function for sorting intervals by location.
inline bool IIbeforeII (const IntervalInfo &i, const IntervalInfo &j)
{
	return (mpfr_cmp (*i. getLeft (), *j. getLeft ()) < 0);
} /* IIbeforeII */

inline void IntervalSet::sort_by_location ()
{
	std::sort (_intervals. begin (), _intervals. end (), IIbeforeII);
	return;
} /* IntervalSet::sort_by_location */

/// A helper function for sorting intervals by width in the reverse order.
inline bool IIbiggerII (const IntervalInfo &i, const IntervalInfo &j)
{
	return (i. getW () > j. getW ());
} /* IIbiggerII */

inline void IntervalSet::sort_by_width ()
{
	std::sort (_intervals. begin (), _intervals. end (), IIbiggerII);
	return;
} /* IntervalSet::sort_by_width */

inline void IntervalSet::condense ()
{
	// prepare a new array of the intervals (for the condesed structure)
	std::vector<IntervalInfo> _new;

	// go through all the intervals and copy them to the new array,
	// while joining adjacent intervals
	auto_mpfr *intervalLeft (0);
	auto_mpfr *intervalRight (0);
	double width;
	int N;
	for (std::vector<IntervalInfo>::iterator it = _intervals. begin ();
		it != _intervals. end (); ++ it)
	{
		// if this is a new interval then remember its start and end
		if (!intervalLeft)
		{
			intervalLeft = (*it). getLeft ();
			intervalRight = (*it). getRight ();
			width = (*it). getW ();
			N = (*it). getN ();
		}

		// if the new interval is a continuation of the previous one
		// then delete the meeting point and continue with the larger one
		else if (mpfr_cmp (*intervalRight, *(it -> getLeft ())) == 0)
		{
			delete intervalRight;
			delete (*it). getLeft ();
			intervalRight = (*it). getRight ();
			width = 0;
			N += (*it). getN ();
		}

		// if the new interval is disjoint from the previous one
		// then store the previous interval and reset to the new one
		else
		{
			_new. push_back (IntervalInfo (N, 0, 0,
				intervalLeft, intervalRight, width));
			intervalLeft = (*it). getLeft ();
			intervalRight = (*it). getRight ();
			width = (*it). getW ();
			N = (*it). getN ();
		}
	}
	if (intervalLeft)
	{
		_new. push_back (IntervalInfo (N, 0, 0,
			intervalLeft, intervalRight, width));
	}

	// make the new (condensed) structure replace the old one
	_intervals. swap (_new);

	return;
} /* IntervalSet::condense */


// --------------------------------------------------
// ------------------- CONTROLLER -------------------
// --------------------------------------------------

/// A class whose object is expected to control the computations:
/// (1) provide intervals of parameters for iterating,
/// (2) make decisions whether to continue after each iteration step,
/// (3) decide what to do if erratic situations happen.
/// This object is also (very) frequently notified about the progress
/// of the computations, so that it can show messages if appropriate,
/// and gather all kinds of statistics.
class Controller
{
public:
	/// The default constructor.
	Controller (const auto_mpfr &leftOmegaFull,
		const auto_mpfr &rightOmegaFull,
		int maxIterations, int maxIntervals, size_t maxQueueSize,
		int maxMessages, int timeLimit, int targetN,
		int requiredN, const auto_mpfr &requiredWidth,
		int showWorstIntervals,
		const auto_mpfr &fraction, const auto_mpfr &minWidth,
		int contractions, int bisectionSteps,
		int outDigits, int verbosity);

	/// The destructor.
	~Controller ();

	/// Adds an interval for starting the computations from.
	/// The provided number indicates the least number
	/// of expected iterates until omega_N hits Delta.
	/// If the interval is too small then it is abandoned.
	int add (const auto_mpfr &leftOmega,
		const auto_mpfr &rightOmega, int N, int D);

	/// Returns the number of intervals waiting in the queue.
	size_t size () const;

	/// Takes an interval out from the queue.
	/// Returns the number of expected iterates, saved previously.
	int get (auto_mpfr &leftOmega, auto_mpfr &rightOmega);

	/// Takes a note of the completion of an iteration step,
	/// and shows messages if necessary.
	void iteration_completed (int N,
		const auto_mpfr &leftNinner, const auto_mpfr &leftNouter,
		const auto_mpfr &rightNinner, const auto_mpfr &rightNouter);

	/// Verifies bound on the derivative to see if the bounding interval
	/// gets closer and closer to the origin in a periodic way.
	/// If this happens then asks the process to stop the iterations
	/// and to abandon the interval.
	/// This is a heuristic procedure, and it might not detect
	/// all the cases in which the iterated interval gets trapped
	/// in a neighborhood of an attracting periodic orbit.
	/// Returns: 0 = no problems, -1 = abandon the interval.
	int check_derivative (int N,
		const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
		const auto_mpfr &leftNouter, const auto_mpfr &rightNouter,
		const auto_mpfr &lowF, const auto_mpfr &uppF);

	/// Verifies if the number of allowed iterates has been reached
	/// and if the amount of time spent on iterating is not too big.
	/// Also verifies if the next interval to iterate (at the top
	/// of the queue) was iterated are least the target number of times.
	/// Returns the information on what to do:
	/// 0 - continue iterating, all OK
	/// nonzero - stop iterating (shows a message on the reason).
	/// It is recommended to call this method with N=-1
	/// each time before taking the next interval to iterate.
	int check_iterating (int N,
		const auto_mpfr &leftOmega, const auto_mpfr &rightOmega);

	/// Verifies if the interval provided is correct.
	/// If the left endpoint is greater than the right endpoint
	/// then shows a message and throws an exception.
	void check_interval (int N, const std::string &name,
		const auto_mpfr &left, const auto_mpfr &right);

	/// Takes a note of a computed interval for some quantity.
	/// Verifies if the interval provided is correct.
	/// Shows the interval if necessary.
	void interval_computed (int N, const std::string &name,
		const auto_mpfr &left, const auto_mpfr &right);

	/// Takes a note of a serious error when computing minus2xy.
	/// Shows a message with the intervals multiplied.
	void minus2xyError (int N, const std::string &name,
		const auto_mpfr &left1, const auto_mpfr &right1,
		const auto_mpfr &left2, const auto_mpfr &right2);

	/// Takes a note of a mismatch in the signs the given quantity
	/// (either "C" or "F") at the endpoints of the interval,
	/// and splits the interval in half.
	/// The number N indicates the iterate at which the problem
	/// with the sign occurred (so the interval was in fact
	/// iterated successfully N-1 times).
	void signMismatch (int N, int reason,
		const auto_mpfr &leftOmega,
		const auto_mpfr &rightOmega);

	/// Takes a note that omega_n has just hit Delta.
	/// Shows a message if necessary.
	/// Returns 1 if satisfied with this result (no more iterations
	/// or splitting necessary, forget this interval),
	/// or 0 if the interval must be split and the iterations
	/// should be continued with smaller parts.
	int omega_hit_delta (int N,
		const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
		const auto_mpfr &delta, const auto_mpfr &minusDelta,
		const auto_mpfr &leftNinner, const auto_mpfr &leftNouter,
		const auto_mpfr &rightNinner, const auto_mpfr &rightNouter,
		const auto_mpfr &lowC, const auto_mpfr &uppC,
		const auto_mpfr &lowF, const auto_mpfr &uppF,
		const auto_mpfr &lowD, const auto_mpfr &uppD,
		bool increasing);

	/// Adds a given part of the large interval (split after having
	/// hit Delta) if the part is big enough, and shows a message.
	void add_interval_part (int N,
		const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
		const auto_mpfr &leftBound, const auto_mpfr &rightBound);

	/// Adds parts of the interval after it had hit Delta.
	/// Shows information and adds the intervals to the queue,
	/// except when the interval is very small.
	void add_interval_parts (int N,
		const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
		const auto_mpfr &leftMiddle, const auto_mpfr &rightMiddle);

	/// Gathers information about parameter exclusions.
	/// Makes a note that at the Nth iteration, the image of the
	/// given interval omega by c_N hit Delta, and therefore
	/// the interval provided will be excluded from further iterations.
	void interval_excluded (int N,
		const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
		const auto_mpfr &leftExcluded,
		const auto_mpfr &rightExcluded);

	/// Gathers information about abandoned intervals of parameters.
	/// Makes a note that at the Nth iteration of the given interval
	/// omega, some problem was found (e.g. the image didn't grow),
	/// so the interval provided will be excluded from further processing.
	/// In particular, updates the measure of all the abandoned intervals.
	/// Shows no messages, so before or after calling this procedure,
	/// it is necessary to display appropriate messages if desired.
	void interval_abandoned (int N,
		const auto_mpfr &leftOmega, const auto_mpfr &rightOmega);

	/// Takes a note of the fact that the inner bound on omega_N
	/// got empty, and splits the interval in half.
	void innerBoundEmpty (int N, const auto_mpfr &leftOmega,
		const auto_mpfr &rightOmega);

	/// Return a recommended number of bisection steps.
	int bisectionSteps ();

	/// Finalizes the coordination: clears the queue
	// and displays a summary of what was done.
	void finalize ();

private:
	/// The copy constructor is not allowed.
	Controller (const Controller &src);

	/// The assignment operator is not allowed.
	Controller &operator = (const Controller &src);

	// ------------------------------------------
	// parameters provided in the constructor:
	// ------------------------------------------

	/// The left bound of the large interval Omega.
	auto_mpfr _leftOmegaFull;

	/// The right bound of the large interval Omega.
	auto_mpfr _rightOmegaFull;

	/// The maximal number of iterates allowed.
	int _maxIterations;

	/// The maximal number of intervals iterated allowed.
	int _maxIntervals;

	/// The maximal number of intervals waiting in the queue.
	size_t _maxQueueSize;

	/// The maximal number of messages to be displayed.
	int _maxMessages;

	/// The maximal number of seconds allowed for the computations.
	int _timeLimit;

	/// The target number of iterations (stop when it is reached).
	int _targetN;

	/// The required minimal N to be satisfied with an interval.
	int _requiredN;

	/// The required minimal width of omega_n to be satisfied.
	auto_mpfr _requiredWidth;

	/// The number of worst intervals to show at the end.
	int _showWorstIntervals;

	/// The minimal required fraction of an interval to iterate further.
	auto_mpfr _fraction;

	/// The minimal width of an interval to be added to the queue.
	auto_mpfr _minWidth;

	/// The allowed number of consecutive contractions detected.
	int _contractions;

	/// The number of bisection steps when excluding a subinterval.
	int _bisectionSteps;

	/// The number of digits to use for outputting high-prec numbers.
	int _outDigits;

	/// The verbosity level for showing messages.
	int _verbosity;

	// ------------------------------------------
	// internal variables that help controlling and gather statistics
	// ------------------------------------------

	/// A counter of intervals added to the queue with "add".
	long _counterAdd;

	/// A counter of intervals taken for iteration with "get".
	long _counterGet;

	/// A counter of the messages displayed.
	long _counterMsg;

	/// An upper bound on the total measure of all the excluded intervals.
	auto_mpfr _excludedMeasure;

	/// The number of excluded intervals.
	long _excludedCount;

	/// An upper bound on the total measure of all abandoned intervals.
	auto_mpfr _abandonedMeasure;

	/// The number of abandoned intervals.
	long _abandonedCount;

	/// A lower bound on the parameter intervals that provide
	/// satisfactory N and width of omega_N.
	auto_mpfr _goodMeasure;

	/// The number of satisfactory parameter intervals.
	long _goodCount;

	/// The highest N encountered during the iterations.
	int _highestN;

	/// The distance of a bound on F' from the origin
	/// in the previous iteration.
	auto_mpfr _prevFabs;

	/// The left point of the image of omega for which F' is stored.
	auto_mpfr _prevLeft;

	/// The right point of the image of omega for which F' is stored.
	auto_mpfr _prevRight;

	/// The number of iterate for which the previous F' is stored.
	int _prevN;

	/// The previous period after which the interval returned.
	int _prevPeriod;

	/// The number of iterates during which the distance of F'
	/// from the origin was decreasing.
	/// Set to -1 if the previous F' is unknown.
	/// Set to 0 if it was previously increasing.
	int _decreasingFabs;

	/// An object used to measure the time used by the program.
	chomp::homology::timeused _compTime;

	// The next time (measured in seconds from the start of the program)
	// to show a periodic summary of the progress of the computations.
	// Updated after the summary is shown.
	int _nextSummaryTime;

	/// Was the message about the reason for quitting the program
	/// already displayed?
	bool _quitAnnounced;

	/// The number of times the inner bound on omega_n hit Delta.
	int _hitDeltaInnerCount;

	/// The number of times the outer bound on omega_n only hit Delta.
	int _hitDeltaOuterCount;

	/// Intervals to be processed as starting points for iterations.
	std::priority_queue<IntervalInfo> _intervals;

	/// Counter of reasons for why the intervals are in the queue.
	ReasonCounter _reasons;

	/// The set of all the excluded and abandoned intervals.
	/// Gathered only when collecting 'escape' intervals.
	/// At the end, the intervals in the queue are added,
	/// the set is sorted, condensed, and the information is displayed.
	IntervalSet _bad;

}; /* class Controller */

// --------------------------------------------------

inline Controller::Controller (const auto_mpfr &leftOmegaFull,
	const auto_mpfr &rightOmegaFull,
	int maxIterations, int maxIntervals,
	size_t maxQueueSize, int maxMessages, int timeLimit, int targetN,
	int requiredN, const auto_mpfr &requiredWidth,
	int showWorstIntervals,
	const auto_mpfr &fraction, const auto_mpfr &minWidth,
	int contractions, int bisectionSteps, int outDigits, int verbosity):
	_maxIterations (maxIterations), _maxIntervals (maxIntervals),
	_maxQueueSize (maxQueueSize), _maxMessages (maxMessages),
	_timeLimit (timeLimit), _targetN (targetN), _requiredN (requiredN),
	_showWorstIntervals (showWorstIntervals),
	_contractions (contractions),
	_bisectionSteps (bisectionSteps),
	_outDigits (outDigits), _verbosity (verbosity),
	_counterAdd (0), _counterGet (0), _counterMsg (0),
	_excludedCount (0), _abandonedCount (0), _goodCount (0),
	_highestN (0),
	_decreasingFabs (-1), _compTime (), _nextSummaryTime (0),
	_quitAnnounced (false),
	_hitDeltaInnerCount (0), _hitDeltaOuterCount (0)
{
	mpfr_set (_leftOmegaFull, leftOmegaFull, GMP_RNDD);
	mpfr_set (_rightOmegaFull, rightOmegaFull, GMP_RNDU);
	mpfr_set (_requiredWidth, requiredWidth, GMP_RNDU);
	mpfr_set (_fraction, fraction, GMP_RNDD);
	mpfr_set (_minWidth, minWidth, GMP_RNDD);
	mpfr_set_ui (_excludedMeasure, 0, GMP_RNDU);
	mpfr_set_ui (_abandonedMeasure, 0, GMP_RNDU);
	mpfr_set_ui (_goodMeasure, 0, GMP_RNDD);
	_compTime = 0;

	// notify the user that small intervals will not be added to the queue
	if ((_verbosity >= Verbosity::abandonedSmall) &&
		(mpfr_sgn (minWidth) > 0))
	{
		using chomp::homology::sout;
		sout << "Note: When adding intervals to the queue, "
			"all intervals whose width is not proved "
			"to be above " <<
			mpfr2string (_minWidth, _outDigits, GMP_RNDD) <<
			" will be abandoned.\n";
	}
	return;
} /* Controller::Controller */

inline Controller::~Controller ()
{
	while (!_intervals. empty ())
	{
		IntervalInfo ii = _intervals. top ();
		delete ii. getLeft ();
		delete ii. getRight ();
		_intervals. pop ();
	}
	return;
} /* Controller::~Controller */

inline Controller::Controller (const Controller &)
{
	throw "The copy constructor not allowed in the Controller class.";
	return;
} /* Controller::Controller */

inline Controller &Controller::operator = (const Controller &)
{
	throw "The assignment operator not allowed in the Controller class.";
	return *this;
} /* Controller::operator = */

inline int Controller::add (const auto_mpfr &leftOmega,
	const auto_mpfr &rightOmega, int N, int D)
{
	// compute a lower bound on the width of the interval
	// to make sure it is not too small
	auto_mpfr width;
	mpfr_sub (width, rightOmega, leftOmega, GMP_RNDD);

	// if the lower bound on the width is below the required width
	// then abandon this interval
	if (mpfr_cmp (width, _minWidth) < 0)
	{
		if (_verbosity >= Verbosity::abandonedSmall)
		{
			using chomp::homology::sout;
			++ _counterMsg;
			sout << "Abandoning an interval contained in [" <<
				mpfr2string (leftOmega, _outDigits,
				GMP_RNDD) << "," <<
				mpfr2string (rightOmega, _outDigits,
				GMP_RNDU) << "], because a lower bound " <<
				mpfr2string (width, _outDigits, GMP_RNDD) <<
				" on its width is too small.\n";
		}

		if (_verbosity >= Verbosity::exclusionData)
		{
			using chomp::homology::sout;
			++ _counterMsg;

			// compute an upper bound on the width
			// of the interval
			auto_mpfr width;
			mpfr_sub (width, rightOmega, leftOmega, GMP_RNDU);

			sout << "S " << N << " " <<
				mpfr2string (leftOmega, _outDigits,
				GMP_RNDD) << " " <<
				mpfr2string (width, _outDigits,
				GMP_RNDU) << "\n";
		}

		// make a note that this interval was abandoned
		this -> interval_abandoned (N, leftOmega, rightOmega);

		return _intervals. size ();
	}

	// increase the counter of intervals added to the queue
	++ _counterAdd;

	// make a copy of the endpoints of the interval
	auto_mpfr *leftNew = new auto_mpfr;
	mpfr_set (*leftNew, leftOmega, GMP_RNDD);
	auto_mpfr *rightNew = new auto_mpfr;
	mpfr_set (*rightNew, rightOmega, GMP_RNDU);

	// add all this information to the queue
	_intervals. push (IntervalInfo (N, _counterAdd, D,
		leftNew, rightNew, mpfr_get_d (width, GMP_RNDN)));
	_reasons. inc (D);

	// show a message about the interval added to the queue
	if (_verbosity >= Verbosity::addToQueue)
	{
		using chomp::homology::sout;
		++ _counterMsg;
		sout << "Adding an interval contained in [" <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDD) <<
			"," <<
			mpfr2string (rightOmega, _outDigits, GMP_RNDU) <<
			"] to the queue.\n";
		sout << "Its width is at least " <<
			mpfr2string (width, _outDigits, GMP_RNDD);
		mpfr_sub (width, rightOmega, leftOmega, GMP_RNDU);
		sout << " and at most " <<
			mpfr2string (width, _outDigits, GMP_RNDU) << ".\n";
		sout << "It was iterated " << N << " times.\n";
		sout << "Its consecutive number is " << _counterAdd << ".\n";
	}

	// return the size of the queue with the numbers
	return _intervals. size ();
} /* Controller::add */

inline size_t Controller::size () const
{
	return _intervals. size ();
} /* Controller::size */

inline int Controller::get (auto_mpfr &leftOmega,
	auto_mpfr &rightOmega)
{
	// increase the counter of intervals taken for iteration
	++ _counterGet;

	// reset the counter of contracions
	_decreasingFabs = -1;

	// degueue the information about the interval
	IntervalInfo ii = _intervals. top ();
	_intervals. pop ();

	// set the endpoints and deallocate memory
	mpfr_set (leftOmega, *ii. getLeft (), GMP_RNDD);
	delete ii. getLeft ();
	mpfr_set (rightOmega, *ii. getRight (), GMP_RNDU);
	delete ii. getRight ();

	// recall the previous number of iterations
	int N = ii. getN ();

	// recall the prvious number of the interval
	int counterOrg = ii. getC ();

	// recall the reason for adding the interval to the queue
	int D = ii. getD ();

	// update the reasons counter
	_reasons. dec (D);

	// show the current interval and a bound on its width
	// at the current iterate
	if (_verbosity >= Verbosity::init)
	{
		using chomp::homology::sout;
		++ _counterMsg;
		sout << "\n*** Computation no. " << _counterGet << " ***\n";
		sout << "Taking the interval omega "
			"contained in [" <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDD) <<
			"," <<
			mpfr2string (rightOmega, _outDigits, GMP_RNDU) <<
			"] for iterations.\n";
		sout << "It was put to the queue as interval number " <<
			counterOrg << ".\n";
		auto_mpfr width;
		mpfr_sub (width, rightOmega, leftOmega, GMP_RNDD);
		sout << "Its width is at least " <<
			mpfr2string (width, _outDigits, GMP_RNDD);
		mpfr_sub (width, rightOmega, leftOmega, GMP_RNDU);
		sout << " and at most " <<
			mpfr2string (width, _outDigits, GMP_RNDU) << ".\n";

		// say how many times it was iterated previously
		if (N)
			sout << "It was previously iterated " <<
				N << ((N > 1) ? " times.\n" : " time.\n");
		else
			sout << "It wasn't iterated yet.\n";

		// show the reason for which the interval was in the queue
		if (D == Reason::init)
			sout << "It was in the queue as a new interval "
				"added at initialization.\n";
		else if (D == Reason::chop)
			sout << "It was in the queue as part "
				"of a chopped interval that hit Delta.\n";
		else if (D == Reason::problemC)
			sout << "It was in the queue as part "
				"of a halved interval (problem with C).\n";
		else if (D == Reason::problemF)
			sout << "It was in the queue as part "
				"of a halved interval (problem with F).\n";
		else
			sout << "It was in the queue "
				"for an unknown reason.\n";
		sout << "There are still " << _reasons. get (D) <<
			" intervals added for the same reason.\n";

		if (_intervals. size () > 1)
		{
			sout << "There are still " << _intervals. size () <<
				" intervals";
		}
		else if (_intervals. size () == 1)
			sout << "There is still 1 interval";
		else
			sout << "There are no other intervals";
		sout << " waiting in the queue.\n";
	}

	// return the size of the queue with the numbers
	return N;
} /* Controller::get */

inline void Controller::iteration_completed (int N,
	const auto_mpfr &leftNinner, const auto_mpfr &leftNouter,
	const auto_mpfr &rightNinner, const auto_mpfr &rightNouter)
{
	// show the current interval and a bound on its width
	// at the current iterate
	if (_verbosity >= Verbosity::iterates)
	{
		using chomp::homology::sout;
		++ _counterMsg;
		sout << "Outer bound on omega_" << N << ": [" <<
			mpfr2string (leftNouter, _outDigits, GMP_RNDD) <<
			"," <<
			mpfr2string (rightNouter, _outDigits, GMP_RNDU) <<
			"].\n";
		auto_mpfr width;
		mpfr_sub (width, rightNinner, leftNinner, GMP_RNDD);
		sout << "Inner width of omega_" << N << " is at least " <<
			mpfr2string (width, _outDigits, GMP_RNDD) << ".\n";
	}

	// show how much the inner and outer bounds of omega_n get apart
	if (_verbosity >= Verbosity::boundsApart)
	{
		using chomp::homology::sout;
		++ _counterMsg;
		auto_mpfr leftWidth;
		mpfr_sub (leftWidth, leftNinner, leftNouter, GMP_RNDU);
		auto_mpfr rightWidth;
		mpfr_sub (rightWidth, rightNouter, rightNinner, GMP_RNDU);
		sout << "The inner and outer bounds " <<
			"for omega_" << N << " got apart by at most " <<
			mpfr2string (leftWidth, _outDigits, GMP_RNDU) <<
			" (left) and " <<
			mpfr2string (rightWidth, _outDigits, GMP_RNDU) <<
			" (right).\n";
	}

	return;
} /* Controller::iteration_completed */

inline int Controller::check_derivative (int N,
	const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
	const auto_mpfr &leftNouter, const auto_mpfr &rightNouter,
	const auto_mpfr &lowF, const auto_mpfr &uppF)
{
	using chomp::homology::sout;

	// if contractions verification is not to be used, do nothing
	if (_contractions <= 0)
		return 0;

	// the maximal period for which we look for returns
	const int MaxPeriod (30);

	// the required closeness of F' to the origin
	const double MaxFabs (0.01);

	// if the number of the iterate is low then don't check anything
	if (N < 20)
		return 0;

	// if F' wasn't known previously then remember the distance
	// of the bound on F' from zero, and the position of the interval,
	// unless the distance is beyond the required closeness
	if (_decreasingFabs < 0)
	{
		if (mpfr_sgn (lowF) > 0)
		{
			if (mpfr_cmp_d (lowF, MaxFabs) > 0)
				return 0;
			mpfr_set (_prevFabs, lowF, GMP_RNDD);
		}
		else if (mpfr_sgn (uppF) < 0)
		{
			if (mpfr_cmp_d (uppF, -MaxFabs) < 0)
				return 0;
			mpfr_abs (_prevFabs, uppF, GMP_RNDD);
		}
		else
			throw "It seems that F' contains zero.";
		mpfr_set (_prevLeft, leftNouter, GMP_RNDD);
		mpfr_set (_prevRight, rightNouter, GMP_RNDU);
		_prevN = N;
		_decreasingFabs = 0;
		_prevPeriod = -1;
		if (_verbosity >= Verbosity::contractionHeuristics)
		{
			++ _counterMsg;
			sout << "Contraction heuristics: Storing the "
				"information for the first time; N = " <<
				N << ", the lower bound on F' is below " <<
				mpfr2string (_prevFabs, _outDigits,
				GMP_RNDD) << ".\n";
		}
		return 0;
	}

	// check if the interval intersects the image stored previously
	bool intersects = ((mpfr_cmp (_prevLeft, rightNouter) < 0) &&
		mpfr_cmp (_prevRight, leftNouter) > 0);
	if (_verbosity >= Verbosity::contractionHeuristics)
	{
		++ _counterMsg;
		sout << "Contraction heuristics: at N = " << N <<
			", the iterate " <<
			(intersects ? "INTERSECTS" : "doesn't intersect") <<
			" the image from N = " << _prevN << ".\n";
	}

	// if it doesn't intersect and less than the prescribed number
	// of iterations took place then don't check anything
	if (!intersects && (N - _prevN < MaxPeriod))
		return 0;

	// store the new interval if it doesn't intersect
	// after so many iterations, and reset the checking
	if (!intersects)
	{
		mpfr_set (_prevLeft, leftNouter, GMP_RNDD);
		mpfr_set (_prevRight, rightNouter, GMP_RNDU);
		_prevN = N;
		_decreasingFabs = 0;
		_prevPeriod = -1;
		if (_verbosity >= Verbosity::contractionHeuristics)
		{
			++ _counterMsg;
			sout << "Contraction heuristics: "
				"No return yet after " <<
				MaxPeriod << "steps. Resetting.\n";
		}
		return 0;
	}

	// check if the return time is the same as it was previously
	if ((_prevPeriod > 0) && (N - _prevN != _prevPeriod))
	{
		mpfr_set (_prevLeft, leftNouter, GMP_RNDD);
		mpfr_set (_prevRight, rightNouter, GMP_RNDU);
		_prevN = N;
		_decreasingFabs = 0;
		_prevPeriod = -1;
		if (_verbosity >= Verbosity::contractionHeuristics)
		{
			++ _counterMsg;
			sout << "Contraction heuristics: "
				"Period not matching: previously " <<
				_prevPeriod << ", now " << (N - _prevN) <<
				". Resetting.\n";
		}
		return 0;
	}

	// check if the distance from the origin decreased or not
	if (mpfr_sgn (lowF) > 0)
	{
		if (mpfr_cmp (_prevFabs, lowF) > 0)
		{
			++ _decreasingFabs;
			if (_prevPeriod < 0)
				_prevPeriod = N - _prevN;
		}
		else
			_decreasingFabs = 0;
		mpfr_set (_prevFabs, lowF, GMP_RNDD);
		_prevN = N;
	}
	else if (mpfr_sgn (uppF) < 0)
	{
		auto_mpfr dist;
		mpfr_abs (dist, uppF, GMP_RNDD);
		if (mpfr_cmp (_prevFabs, dist) > 0)
		{
			++ _decreasingFabs;
			if (_prevPeriod < 0)
				_prevPeriod = N - _prevN;
		}
		else
			_decreasingFabs = 0;
		mpfr_set (_prevFabs, dist, GMP_RNDD);
		_prevN = N;
	}
	else
		throw "It seems that F' contains zero.";

	if (_verbosity >= Verbosity::contractionHeuristics)
	{
		++ _counterMsg;
		sout << "Contraction heuristics: decrease no. " <<
			_decreasingFabs << " observed, "
			"the returning period is " <<
			_prevPeriod << ".\n";
	}

	// if the distance decreased the number of times
	// that is still allowed then don't do anything
	if (_decreasingFabs < _contractions)
		return 0;

	// make a note that this interval was abandoned
	this -> interval_abandoned (N, leftOmega, rightOmega);

	// show a message and request abandoning the interval
	if (_verbosity >= Verbosity::abandonedAttracting)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		// compute an upper bound on the width
		// of the original interval that was iterated
		auto_mpfr width;
		mpfr_sub (width, rightOmega, leftOmega, GMP_RNDU);

		sout << "The distance of a bound on F' from the origin "
			"was decreasing for at least " << _decreasingFabs <<
			" iterations, and now F_" << N <<
			" is not greater than " <<
			mpfr2string (_prevFabs, _outDigits, GMP_RNDU) << ".\n"
			"Abandoning the interval [" <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDD) <<
			"," <<
			mpfr2string (rightOmega, _outDigits, GMP_RNDU) <<
			"]. Its width is at most " <<
			mpfr2string (width, _outDigits, GMP_RNDU) <<
			"\n";
	}

	if (_verbosity >= Verbosity::exclusionData)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		// compute an upper bound on the width
		// of the original interval that was iterated
		auto_mpfr width;
		mpfr_sub (width, rightOmega, leftOmega, GMP_RNDU);

		sout << "T " << N << " " <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDD) <<
			" " <<
			mpfr2string (width, _outDigits, GMP_RNDU) << "\n";
	}

	return -1;
} /* Controller::check_derivative */

inline int Controller::check_iterating (int N,
	const auto_mpfr &leftOmega, const auto_mpfr &rightOmega)
{
	// check if the limit on the number of iterations has been exceeded
	if (N >= _maxIterations)
	{
		using chomp::homology::sout;

		if (_verbosity >= Verbosity::abandonedLong)
		{
			using chomp::homology::sout;
			++ _counterMsg;
			sout << "Abandoning an interval contained in [" <<
				mpfr2string (leftOmega, _outDigits,
				GMP_RNDD) << "," <<
				mpfr2string (rightOmega, _outDigits,
				GMP_RNDU) <<
				"], because it was iterated more than " <<
				_maxIterations << " times.\n";
		}

		if ((_verbosity >= Verbosity::iteratedData) ||
			(_verbosity >= Verbosity::exclusionData))
		{
			using chomp::homology::sout;
			++ _counterMsg;

			// compute an upper bound on the width
			// of the interval
			auto_mpfr width;
			mpfr_sub (width, rightOmega, leftOmega, GMP_RNDU);

			sout << "A " << N << " " <<
				mpfr2string (leftOmega, _outDigits,
				GMP_RNDD) << " " <<
				mpfr2string (width, _outDigits,
				GMP_RNDU) << "\n";
		}

		// make a note that this interval was abandoned
		this -> interval_abandoned (N, leftOmega, rightOmega);

		return 1;
	}

	// update the counter of the highest number of iterations encountered
	if (N > _highestN)
		_highestN = N;

	// check if the number of iterated intervals has been reached
	if ((N < 0) && (_maxIntervals > 0) && (_counterGet >= _maxIntervals))
	{
		if (_quitAnnounced)
			return 2;
		using chomp::homology::sout;
		++ _counterMsg;
		if (_verbosity >= Verbosity::stopReason)
		{
			sout << "STOP! The number of " << _maxIntervals <<
				" intervals to process reached. "
				"Interrupting the program.\n\n";
		}
		_quitAnnounced = true;
		return 1;
	}

	// check if the max number of intervals in the queue has been reached
	if ((_maxQueueSize > 0) && (_intervals. size () > _maxQueueSize))
	{
		if (_quitAnnounced)
			return 2;
		using chomp::homology::sout;
		++ _counterMsg;
		if (_verbosity >= Verbosity::stopReason)
		{
			sout << "STOP! The number of " << _maxQueueSize <<
				" intervals waiting in the queue exceeded. "
				"Interrupting the program.\n\n";
		}
		_quitAnnounced = true;
		return 1;
	}

	// determine the time elapsed so far
	double elapsedTime = static_cast<double> (_compTime);

	// check if the time limit has been exceeded
	if ((_timeLimit >= 0) && (elapsedTime > _timeLimit))
	{
		if (_quitAnnounced)
			return 2;
		using chomp::homology::sout;
		++ _counterMsg;
		if (_verbosity >= Verbosity::stopReason)
		{
			sout << "STOP! The time limit of " << _timeLimit <<
				" sec exceeded. "
				"Interrupting the program.\n\n";
		}
		_quitAnnounced = true;
		return 1;
	}

	// check if the max number of intervals in the queue has been reached
	if ((_maxMessages > 0) && (_counterMsg > _maxMessages))
	{
		if (_quitAnnounced)
			return 2;
		using chomp::homology::sout;
		++ _counterMsg;
		if (_verbosity >= Verbosity::stopReason)
		{
			sout << "STOP! The number of " << _maxMessages <<
				" messages to display exceeded. "
				"Interrupting the program.\n\n";
		}
		_quitAnnounced = true;
		return 1;
	}

	// check if the target number of iterates has been reached
	if (_targetN && (N < 0) && !_intervals. empty () &&
		(_intervals. top (). getN () >= _targetN))
	{
		using chomp::homology::sout;
		if (_verbosity >= Verbosity::stopReason)
		{
			sout << "STOP! The target number of " << _targetN <<
				" iterations has been reached by all the intervals "
				"in the queue.\n\n";
		}
		return 1;
	}

	// show a periodic summary of the results if necessary
	if ((N < 0) && (_verbosity >= Verbosity::periodicSummary) &&
		(elapsedTime >= _nextSummaryTime))
	{
		using chomp::homology::sout;
		++ _counterMsg;
		sout << "Info " << _nextSummaryTime << ": ";
		sout << _counterGet << " iterated, ";
		sout << _intervals. size () << " in the queue, ";
		sout << "N = " << (_intervals. empty () ? 0 :
			 _intervals. top (). getN ()) << " minimum, ";
		if (_requiredN > 0)
		{
			sout << _goodCount << " good: " <<
				mpfr2string (_goodMeasure, _outDigits,
				GMP_RNDD) << ", ";
		}
		sout << _excludedCount << " excluded: " <<
			mpfr2string (_excludedMeasure, _outDigits,
			GMP_RNDU) << ", ";
		sout << _abandonedCount << " abandoned: " <<
			mpfr2string (_abandonedMeasure, _outDigits,
			GMP_RNDU) << ".\n";
		++ _nextSummaryTime;
	}

	return 0;
} /* Controller::check_iterating */

inline void Controller::check_interval (int N, const std::string &name,
	const auto_mpfr &left, const auto_mpfr &right)
{
	// check if the bounds define a correct interval
	if (mpfr_cmp (left, right) > 0)
	{
		using chomp::homology::sout;
		++ _counterMsg;
		sout << "The interval bounds on " << name << "_" << N <<
			" are wrong: [" <<
			mpfr2string (left, _outDigits, GMP_RNDU) << "," <<
			mpfr2string (right, _outDigits, GMP_RNDD) << "].\n";
		throw "An wrong interval detected: left bound > right bound.";
	}

	return;
} /* Controller::check_interval */

inline void Controller::interval_computed (int N, const std::string &name,
	const auto_mpfr &left, const auto_mpfr &right)
{
	// first verify the interval if it is correct
	check_interval (N, name, left, right);

	// show the interval computed if the verbosity level is so high
	if (_verbosity >= Verbosity::allComputedIntervals)
	{
		using chomp::homology::sout;
		++ _counterMsg;
		sout << "The interval " << name << "_" << N <<
			" is contained in [" <<
			mpfr2string (left, _outDigits, GMP_RNDD) << "," <<
			mpfr2string (right, _outDigits, GMP_RNDU) << "].\n";
	}

	// update the counter of the highest number of iterations encountered
	if (N > _highestN)
		_highestN = N;

	return;
} /* Controller::interval_computed */

inline void Controller::minus2xyError (int N, const std::string &name,
	const auto_mpfr &left1, const auto_mpfr &right1,
	const auto_mpfr &left2, const auto_mpfr &right2)
{
	using chomp::homology::sout;

	++ _counterMsg;
	sout << "Numerical error: wrong arguments to 'minus2xy' "
		"at iteration " << N << ":\n";
	sout << name << " is contained in [" <<
		mpfr2string (left1, _outDigits, GMP_RNDD) << "," <<
		mpfr2string (right1, _outDigits, GMP_RNDU) << "].\n";
	sout << "The other argument is contained in [" <<
		mpfr2string (left2, _outDigits, GMP_RNDD) << "," <<
		mpfr2string (right2, _outDigits, GMP_RNDU) << "].\n";

	return;
} /* Controller::minus2xyError */

inline void Controller::signMismatch (int N, int reason,
	const auto_mpfr &leftOmega, const auto_mpfr &rightOmega)
{
	if (_verbosity >= Verbosity::iteratedData)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		sout << ((reason == Reason::problemC) ? "C " : "F ") <<
			N << " " <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDN) <<
			"\n";
	}

	if (_verbosity >= Verbosity::signMismatch)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		sout << "A mismatch in the signs of the endpoints of " <<
			((reason == Reason::problemC) ? "C" : "F") <<
			" was detected at iteration " << N <<
			" that started from omega contained in [" <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDD) <<
			"," <<
			mpfr2string (rightOmega, _outDigits, GMP_RNDU) <<
			"]. Splitting the interval in half.\n";
	}

	// compute a middle point between the two endpoints
	auto_mpfr middle;
	mpfr_add (middle, leftOmega, rightOmega, GMP_RNDN);
	mpfr_div_ui (middle, middle, 2, GMP_RNDN);

	// add both halves of the interval
	this -> add (leftOmega, middle, N - 1, reason);
	this -> add (middle, rightOmega, N - 1, reason);

	return;
} /* Controller::signMismatch */

inline int Controller::omega_hit_delta (int N,
	const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
	const auto_mpfr &delta, const auto_mpfr &minusDelta,
	const auto_mpfr &leftNinner, const auto_mpfr &leftNouter,
	const auto_mpfr &rightNinner, const auto_mpfr &rightNouter,
	const auto_mpfr &lowC, const auto_mpfr &uppC,
	const auto_mpfr &lowF, const auto_mpfr &uppF,
	const auto_mpfr &lowD, const auto_mpfr &uppD,
	bool increasing)
{
	// check if omega hit Delta for sure or just "possibly..."
	bool hitForSure = (mpfr_cmp (rightNinner, minusDelta) > 0) &&
		(mpfr_cmp (leftNinner, delta) < 0);
	if (hitForSure)
		++ (this -> _hitDeltaInnerCount);
	else
		++ (this -> _hitDeltaOuterCount);

	// show a message on whether the interval indeed intersects Delta
	if (_verbosity >= Verbosity::omegaHitDelta)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		// if it intersects indeed
		if (hitForSure)
		{
			sout << "Omega_" << N << " hit Delta "
				"(its inner bound intersects Delta).\n";
		}
		// otherwise it was not proven, but is still possible
		else
		{
			sout << "Omega_" << N << " probably hit Delta "
				"(it is only its outer bound "
				"that intersects Delta).\n";
		}

		// show the outer bound on the interval that intersects Delta
		sout << "An outer bound on omega_" << N << ": [" <<
			mpfr2string (leftNouter, _outDigits, GMP_RNDD) <<
			"," <<
			mpfr2string (rightNouter, _outDigits, GMP_RNDU) <<
			"].\n";

		// show the inner width of the interval
		auto_mpfr width;
		mpfr_sub (width, rightNinner, leftNinner, GMP_RNDD);
		sout << "The inner width of omega_" << N << " is at least " <<
			mpfr2string (width, _outDigits, GMP_RNDD) << ".\n";

		// show whether the function c_n is increasing or decreasing
		sout << "The function C_" << N << " is " <<
			(increasing ? "increasing" : "decreasing") <<
			" on omega.\n";
	}

	// if both N and the width of omega_n are satisfactory
	// then make a note and return 1 to inform
	// that we are done with this interval
	if ((_requiredN > 0) && (N >= _requiredN))
	{
		// compute the inner width of the interval omega_N
		auto_mpfr widthN;
		mpfr_sub (widthN, rightNinner, leftNinner, GMP_RNDD);

		if (mpfr_cmp (widthN, _requiredWidth) >= 0)
		{
			// compute a lower bound on the width of omega
			auto_mpfr widthOmega;
			mpfr_sub (widthOmega, rightOmega, leftOmega,
				GMP_RNDD);

			// update the measure of all the good parameters
			mpfr_add (_goodMeasure, _goodMeasure, widthOmega,
				GMP_RNDD);
			++ _goodCount;

			if (_verbosity >= Verbosity::omegaGood)
			{
				using chomp::homology::sout;
				++ _counterMsg;
				sout << "A satisfactory interval found: " <<
					mpfr2string (leftOmega, _outDigits,
					GMP_RNDD) << "," <<
					mpfr2string (rightOmega, _outDigits,
					GMP_RNDU) << ", N = " << N << ".\n";
			}

			if (_verbosity >= Verbosity::goodData)
			{
				using chomp::homology::sout;
				++ _counterMsg;
				sout << "G " << N << " " <<
					mpfr2string (widthN, _outDigits,
					GMP_RNDD) << " " <<
					mpfr2string (leftOmega, _outDigits,
					GMP_RNDD) << " " <<
					mpfr2string (rightOmega, _outDigits,
					GMP_RNDU) << " " <<
					mpfr2string (widthOmega, _outDigits,
					GMP_RNDD) << " " <<
					mpfr2string (lowC, _outDigits,
					GMP_RNDD) << " " <<
					mpfr2string (uppC, _outDigits,
					GMP_RNDU) << " " <<
					mpfr2string (lowF, _outDigits,
					GMP_RNDD) << " " <<
					mpfr2string (uppF, _outDigits,
					GMP_RNDU) << " " <<
					mpfr2string (lowD, _outDigits,
					GMP_RNDD) << " " <<
					mpfr2string (uppD, _outDigits,
					GMP_RNDU) << "\n";
			}

			return 1;
		}
	}

	if (_verbosity >= Verbosity::iteratedData)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		// compute the inner width of the interval omega_N
		auto_mpfr widthN;
		mpfr_sub (widthN, rightNinner, leftNinner, GMP_RNDD);

		sout << "D " << N << " " <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDN) <<
			" " <<
			mpfr2string (widthN, _outDigits, GMP_RNDD) <<
			"\n";
	}

	// show a data line (for statistics)
	if (_verbosity >= Verbosity::hitData)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		// compute the inner width of the interval omega_N
		auto_mpfr widthN;
		mpfr_sub (widthN, rightNinner, leftNinner, GMP_RNDD);

		// compute an upper bound on the width of the interval omega
		auto_mpfr widthOmega;
		mpfr_sub (widthOmega, rightOmega, leftOmega, GMP_RNDU);

		// show the data line
		sout << "H " << N << " " <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDD) <<
			" " <<
			mpfr2string (rightOmega, _outDigits, GMP_RNDU) <<
			" " <<
			mpfr2string (widthOmega, _outDigits, GMP_RNDU) <<
			" " <<
			mpfr2string (leftNouter, _outDigits, GMP_RNDD) <<
			" " <<
			mpfr2string (rightNouter, _outDigits, GMP_RNDU) <<
			" " <<
			mpfr2string (widthN, _outDigits, GMP_RNDD) << "\n";
	}

	// show a data line (for statistics)
	if (_verbosity >= Verbosity::derivativeData)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		// show the data line
		sout << "U " << N << " " <<
			mpfr2string (lowC, _outDigits, GMP_RNDD) << " " <<
			mpfr2string (uppC, _outDigits, GMP_RNDU) << " " <<
			mpfr2string (lowF, _outDigits, GMP_RNDD) << " " <<
			mpfr2string (uppF, _outDigits, GMP_RNDU) << " " <<
			mpfr2string (lowD, _outDigits, GMP_RNDD) << " " <<
			mpfr2string (uppD, _outDigits, GMP_RNDU) << "\n";
	}

	return 0;
} /* Controller::omega_hit_delta */

inline void Controller::add_interval_part (int N,
	const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
	const auto_mpfr &leftBound, const auto_mpfr &rightBound)
{
	// don't do anything if the interval would be empty
	if (mpfr_cmp (leftBound, rightBound) >= 0)
		return;

	// if there is no requirement on the minimal width of the interval
	if (mpfr_sgn (_fraction) == 0)
	{
		// add the part of the interval
		this -> add (leftBound, rightBound, N, Reason::chop);
		return;
	}

	// compute the width of the interval
	auto_mpfr width;
	mpfr_sub (width, rightBound, leftBound, GMP_RNDU);

	// compute the width of the original interval that was iterated
	auto_mpfr widthFull;
	mpfr_sub (widthFull, rightOmega, leftOmega, GMP_RNDD);

	// compute the minimal required width of the interval to be added
	auto_mpfr minWidthAllowed;
	mpfr_mul (minWidthAllowed, widthFull, _fraction, GMP_RNDD);

	// check if the interval is large enough to be added
	if (mpfr_cmp (width, minWidthAllowed) >= 0)
	{
		// add the part of the interval
		this -> add (leftBound, rightBound, N, Reason::chop);
		return;
	}

	// make a note that this interval was abandoned
	this -> interval_abandoned (N, leftOmega, rightOmega);

	// inform about the fact that the interval is too small
	if (_verbosity >= Verbosity::omegaSplit)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		auto_mpfr fraction;
		mpfr_div (fraction, width, widthFull, GMP_RNDU);
		sout << "Abandoning an interval contained in " <<
			mpfr2string (leftBound, _outDigits, GMP_RNDD) <<
			"," <<
			mpfr2string (rightBound, _outDigits, GMP_RNDU) <<
			"], because it is too small: its width is at most " <<
			mpfr2string (width, _outDigits, GMP_RNDU) <<
			", which is no more than " <<
			mpfr2string (fraction, _outDigits, GMP_RNDU) <<
			" of the width of the interval that was split "
			"due to collision with Delta.\n";
	}

	if (_verbosity >= Verbosity::exclusionData)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		sout << "P " << N << " " <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDD) <<
			" " <<
			mpfr2string (width, _outDigits, GMP_RNDU) << "\n";
	}

	return;
} /* Controller::add_interval_part */

inline void Controller::add_interval_parts (int N,
	const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
	const auto_mpfr &leftMiddle, const auto_mpfr &rightMiddle)
{
	using chomp::homology::sout;

	// if the computed interval is wrong then stop the program
	if (mpfr_cmp (leftMiddle, rightMiddle) > 0)
	{
		this -> interval_computed (N, "Middle",
			leftMiddle, rightMiddle);
	}

	// add the left-hand-side part of the interval if it is large enough
	add_interval_part (N, leftOmega, rightOmega, leftOmega, leftMiddle);

	// add the right-hand-side part of the interval if it is large enough
	add_interval_part (N, leftOmega, rightOmega, rightMiddle, rightOmega);

	return;
} /* Controller::add_interval_parts */

inline void Controller::interval_excluded (int N,
	const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
	const auto_mpfr &leftExcluded, const auto_mpfr &rightExcluded)
{
	// compute an upper bound on the width of the excluded interval
	auto_mpfr width;
	mpfr_sub (width, rightExcluded, leftExcluded, GMP_RNDU);

	// add this width to the total measure of excluded intervals
	mpfr_add (_excludedMeasure, _excludedMeasure, width, GMP_RNDU);
	++ _excludedCount;

	// add this interval to the collection of 'bad' intervals if necessary
	if (_showWorstIntervals > 0)
	{
		_bad. add (leftExcluded, rightExcluded,
			mpfr_get_d (width, GMP_RNDN));
	}

	// show a message if necessary
	if (_verbosity >= Verbosity::exclusions)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		// compare the width of the interval with the entire omega
		auto_mpfr fullWidth;
		mpfr_sub (fullWidth, rightOmega, leftOmega, GMP_RNDD);
		auto_mpfr fraction;
		mpfr_div (fraction, width, fullWidth, GMP_RNDU);

		sout << "Excluding an interval contained in [" << 
			mpfr2string (leftExcluded, _outDigits, GMP_RNDD) <<
			"," <<
			mpfr2string (rightExcluded, _outDigits, GMP_RNDU) <<
			"] from an interval contained in [" <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDD) <<
			"," <<
			mpfr2string (rightOmega, _outDigits, GMP_RNDU) <<
			"] because of collision with Delta at c_" << N <<
			".\n";
		sout << "The width of the excluded interval is at most " <<
			mpfr2string (width, _outDigits, GMP_RNDU) <<
			", which is at most " <<
			mpfr2string (fraction, _outDigits, GMP_RNDU) <<
			" of the width of the iterated interval.\n";
		sout << "The total measure of parameters excluded until now "
			"is at most " <<
			mpfr2string (_excludedMeasure, _outDigits,
			GMP_RNDU) << ".\n";
	}

	if (_verbosity >= Verbosity::exclusionData)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		// compare the width of the interval with the entire omega
		auto_mpfr fullWidth;
		mpfr_sub (fullWidth, rightOmega, leftOmega, GMP_RNDD);
		auto_mpfr fraction;
		mpfr_div (fraction, width, fullWidth, GMP_RNDU);

		sout << "E " << N << " " <<
			mpfr2string (leftExcluded, _outDigits, GMP_RNDD) <<
			" " <<
			mpfr2string (width, _outDigits, GMP_RNDU) << "\n";
	}

	return;
} /* Controller::interval_excluded */

inline void Controller::interval_abandoned (int /*N*/,
	const auto_mpfr &leftOmega, const auto_mpfr &rightOmega)
{
	// compute an upper bound on the width of the interval
	auto_mpfr width;
	mpfr_sub (width, rightOmega, leftOmega, GMP_RNDU);

	// add this bound to an upper bound on the measure
	// of all the abandoned intervals
	mpfr_add (_abandonedMeasure, _abandonedMeasure, width, GMP_RNDU);
	++ _abandonedCount;

	// add this interval to the collection of 'bad' intervals if necessary
	if (_showWorstIntervals > 0)
	{
		_bad. add (leftOmega, rightOmega,
			mpfr_get_d (width, GMP_RNDN));
	}

	return;
} /* Controller::interval_abandoned */

inline void Controller::innerBoundEmpty (int N,
	const auto_mpfr &leftOmega, const auto_mpfr &rightOmega)
{
	if (_verbosity >= Verbosity::innerBoundEmpty)
	{
		using chomp::homology::sout;
		++ _counterMsg;

		sout << "The inner bound on omega_" << N <<
			" was found to be empty, omega = [" <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDD) <<
			"," <<
			mpfr2string (rightOmega, _outDigits, GMP_RNDU) <<
			"]. Abandoning it.\n";
	}

	if ((_verbosity >= Verbosity::iteratedData) ||
		(_verbosity >= Verbosity::exclusionData))
	{
		using chomp::homology::sout;
		++ _counterMsg;

		// compute an upper bound on the width
		// of the interval
		auto_mpfr width;
		mpfr_sub (width, rightOmega, leftOmega, GMP_RNDU);

		sout << "I " << N << " " <<
			mpfr2string (leftOmega, _outDigits, GMP_RNDN) <<
			" " <<
			mpfr2string (width, _outDigits, GMP_RNDU) << "\n";
	}

	interval_abandoned (N, leftOmega, rightOmega);

	return;
} /* Controller::innerBoundEmpty */

inline int Controller::bisectionSteps ()
{
	return _bisectionSteps;
} /* Controller::bisectionSteps */

inline void Controller::finalize ()
{
	using chomp::homology::sout;

	if (_verbosity >= Verbosity::computationSummary)
	{
		// show the counters
		sout << "A total of " << _counterMsg <<
			" messages were displayed.\n";
		sout << "A total of " << _counterAdd <<
			" intervals were added to the queue.\n";
		sout << "A total of " << _counterGet <<
			" intervals were taken for iterations.\n";
		sout << "A total of " << _hitDeltaInnerCount <<
			" inner bounds of omega_n hit Delta.\n";
		sout << "A total of " << _hitDeltaOuterCount <<
			" outer bounds of omega_n (but not inner) "
			"hit Delta.\n";

		// show the highest N
		sout << "The highest N encountered was " <<
			_highestN << ".\n";

		sout << "\n";
	}

	// TODO: show how many intervals are in the queue
	// for each of the reasons
	// TODO: compute the measures of the intervals
	// depending on the reasons

	// prepare information about the min and max number of iterates
	// for the queue
	int minN (_intervals. empty () ? 0 : _intervals. top (). getN ());
	int maxN (_intervals. empty () ? 0 : _intervals. top (). getN ());

	// get ready to store the total measure of paramter intervals
	// that are still present in the queue
	auto_mpfr measureMin;
	mpfr_set_ui (measureMin, 0, GMP_RNDD);
	auto_mpfr measureMax;
	mpfr_set_ui (measureMax, 0, GMP_RNDU);

	// remember the size of the queue
	size_t intervalsSize (_intervals. size ());

	// show information about this element if required
	if (intervalsSize && (_verbosity >= Verbosity::leftInQueue))
	{
		sout << "Outer bounds for the " << intervalsSize <<
			" intervals left in the queue, "
			"and some other information:\n";
	}

	// extract all the intervals from the queue, show them,
	// and either forget or move to the list of bad intervals
	while (!_intervals. empty ())
	{
		// extract an element from the beginning of the queue
		IntervalInfo ii = _intervals. top ();
		_intervals. pop ();

		// update min and max N
		int N = ii. getN ();
		if (minN > N)
			minN = N;
		if (maxN < N)
			maxN = N;

		// add the measure of this element
		auto_mpfr widthMin;
		mpfr_sub (widthMin, *ii. getRight (), *ii. getLeft (),
			GMP_RNDD);
		mpfr_add (measureMin, measureMin, widthMin, GMP_RNDD);
		auto_mpfr widthMax;
		mpfr_sub (widthMax, *ii. getRight (), *ii. getLeft (),
			GMP_RNDU);
		mpfr_add (measureMax, measureMax, widthMax, GMP_RNDU);

		// show information about this element if required
		if (_verbosity >= Verbosity::leftInQueue)
		{
			const char *reasons [] = {"initial", "hitDelta",
				"problemC", "problemD"};
			const char *reason = (ii. getD () < 4) ?
				reasons [ii. getD ()] : "unknown";
			sout << "[" <<
				mpfr2string (*ii. getLeft (), _outDigits,
				GMP_RNDD) << "," <<
				mpfr2string (*ii. getRight (), _outDigits,
				GMP_RNDU) << "], width at most " <<
				mpfr2string (widthMax, _outDigits,
				GMP_RNDU) << ", iterated " << ii. getN () <<
				" times, reason added: " << reason << ".\n";
		}

		// move the processed interval to the set of bad intervals,
		// or clear the memory that was used by the interval
		if (_showWorstIntervals > 0)
		{
			_bad. add (ii. getLeft (), ii. getRight (),
				ii. getW ());
		}
		else
		{
			delete ii. getLeft ();
			delete ii. getRight ();
		}
	}
	if (intervalsSize && (_verbosity >= Verbosity::leftInQueue))
		sout << "\n";

	if (_verbosity >= Verbosity::queueSummary)
	{
		// show the queue size and the number of iterations
		if (intervalsSize == 0)
			sout << "No intervals were left in the queue.\n";
		else if (intervalsSize == 1)
		{
			sout << "There was one interval left in the queue.\n";
			sout << "It was iterated " << minN << " times.\n";
		}
		else
		{
			sout << "There were " << intervalsSize <<
				" intervals left in the queue.\n";
			sout << "They were iterated between " << minN <<
				" and " << maxN << " times.\n";
		}

		// show the reason for which the intervals were in the queue
		for (int r = 0; r < static_cast<int> (_reasons. size ());
			++ r)
		{
			int count_reason (_reasons. get (r));
			if (count_reason <= 0)
				continue;
			sout << count_reason << " interval" <<
				((count_reason == 1) ? "" : "s") <<
				" left in the queue " <<
				((count_reason == 1) ? "was" : "were") <<
				" added ";
			if (r == Reason::init)
				sout << "at initialization.\n";
			else if (r == Reason::chop)
				sout << "as part of an interval "
					"that hit Delta.\n";
			else if (r == Reason::problemC)
				sout << "due to a problem with C.\n";
			else if (r == Reason::problemF)
				sout << "due to a problem with F.\n";
			else
				sout << "for an unknown reason.\n";
		}

		// show the measure of all the intervals in the queue
		sout << "The total measure of all the intervals "
			"left in the queue is between " <<
			mpfr2string (measureMin, _outDigits, GMP_RNDD) <<
			" and " <<
			mpfr2string (measureMax, _outDigits, GMP_RNDU) <<
			".\n";

		sout << "\n";
	}

	// compute the total number of remaining intervals and their measure
	int _remainingCount = intervalsSize + _excludedCount +
		_abandonedCount;
	auto_mpfr _remainingMeasure;
	mpfr_set (_remainingMeasure, measureMax, GMP_RNDU);
	mpfr_add (_remainingMeasure, _remainingMeasure, _excludedMeasure,
		GMP_RNDU);
	mpfr_add (_remainingMeasure, _remainingMeasure, _abandonedMeasure,
		GMP_RNDU);

	// compute the approximate percentage of good intervals found
	auto_mpfr fullWidth;
	mpfr_sub (fullWidth, _rightOmegaFull, _leftOmegaFull, GMP_RNDU);
	double approxFullWidth (mpfr_get_d (fullWidth, GMP_RNDU));
	double goodPercentage ((approxFullWidth > 0) ? 100 *
		mpfr_get_d (_goodMeasure, GMP_RNDD) / approxFullWidth : 0);

	// show the basic summary - short and clear
	if ((_requiredN > 0) && (_verbosity >= Verbosity::basicSummary))
	{
		sout << goodPercentage << "% parameters were found to be "
			"in stochastic intervals.\n\n";
	}

	// show the measure of good parameter intervals (if it was computed)
	if ((_requiredN > 0) && (_verbosity >= Verbosity::measureSummary))
	{
		sout << "The number of stochastic parameter intervals "
			"found is " << _goodCount << ".\n";
		sout << "The total measure of the stochastic parameter "
			"intervals is at least " <<
			mpfr2string (_goodMeasure, _outDigits, GMP_RNDD) <<
			".\n";
	}

	if ((_requiredN <= 0) && (_verbosity >= Verbosity::measureSummary))
	{
		sout << "No required number of iterations was defined, "
			"so no stochastic intervals were found.\n";
	}

	// show the total number of remaining intervals and their measure
	if (_verbosity >= Verbosity::measureSummary)
	{
		sout << "All the remaining intervals were either left "
			"in the queue, excluded, or abandoned.\n";
		sout << "The number of the remaining parameter intervals "
			"is " << _remainingCount << ".\n";
		sout << "The total measure of the remaining parameter "
			"intervals is at most " <<
			mpfr2string (_remainingMeasure, _outDigits,
			GMP_RNDU) << ".\n";
	}

	if (_verbosity >= Verbosity::computationSummary)
	{
		// show the total number of intervals that appeared
		sout << "The total number of intervals created is " <<
			(_goodCount + _remainingCount) << ".\n";

		// show the excluded measure
		sout << "The total measure of " << _excludedCount <<
			" excluded parameter intervals is at most " <<
			mpfr2string (_excludedMeasure, _outDigits,
			GMP_RNDU) << ".\n";

		// show the abandoned measure
		sout << "The total measure of " << _abandonedCount <<
			" parameter intervals abandoned "
			"due to various reasons is at most " <<
			mpfr2string (_abandonedMeasure, _outDigits,
			GMP_RNDU) << ".\n";
	}

	// analyze the bad intervals and show a few major ones
	if (_showWorstIntervals > 0)
	{
		// sort the intervals by location, merge adjacent ones,
		// and sort the intervals by width (largest first)
		int nPrev = _bad. size ();
		_bad. sort_by_location ();
		_bad. condense ();
		_bad. sort_by_width ();

		// show the first few intervals
		int nBad = _bad. size ();
		int maxBad = _showWorstIntervals;
		sout << nPrev << " bad intervals merged into " << nBad <<
			" intervals and sorted by width.\n";
		int sumN (0);
		for (int i = 0; (i < maxBad) && (i < nBad); ++ i)
		{
			auto_mpfr *leftOmega (0);
			auto_mpfr *rightOmega (0);
			double width (0);
			int N (0);
			_bad. get (i, leftOmega, rightOmega, &width, &N);
			sumN += N;
			sout << "Bad interval no. " << (i + 1) <<
				" is contained in [" <<
				mpfr2string (*leftOmega, _outDigits,
				GMP_RNDD) << ", " <<
				mpfr2string (*rightOmega, _outDigits,
				GMP_RNDU) << "], width = " << width <<
				". It was built of " << N << " intervals.\n";
		}
		sout << "These bad intervals were built of " << sumN <<
			" smaller intervals in total.\n";
	}

	return;
} /* Controller::finalize */


// --------------------------------------------------
// ---------------- BASIC OPERATIONS ----------------
// --------------------------------------------------

/// Computes a lower bound of the image of a given point
/// under the quadratic map fa(x)=a-x^2.
/// The number "x" is the argument, the number "a" is the parameter.
inline void quadr_lower (auto_mpfr &result,
	const auto_mpfr &arg, const auto_mpfr &param)
{
	auto_mpfr square;
	mpfr_sqr (square, arg, GMP_RNDU);
	mpfr_sub (result, param, square, GMP_RNDD);
	return;
} /* quadr_lower */

/// Computes an upper bound of the image of a given point
/// under the quadratic map fa(x)=a-x^2.
/// The number "x" is the argument, the number "a" is the parameter.
inline void quadr_upper (auto_mpfr &result,
	const auto_mpfr &arg, const auto_mpfr &param)
{
	auto_mpfr square;
	mpfr_sqr (square, arg, GMP_RNDD);
	mpfr_sub (result, param, square, GMP_RNDU);
	return;
} /* quadr_upper */

/// Computes a lower bound on the preimage of a given point
/// under the quadratic map fa(x)=a-x^2, that is, computes
/// x = +/- square root of (a - y).
/// The number "y" is the argument, the number "a" is the parameter,
/// the integer indicates which part of the domain to use for the preimage:
/// 0 means to the left from the critical point, 1 -- to the right.
/// Throws an exception in case this is not possible.
inline void quadr_pre_lower (auto_mpfr &result,
	const auto_mpfr &arg, const auto_mpfr &param, int which)
{
	auto_mpfr diff;
	mpfr_sub (diff, param, arg, which ? GMP_RNDD : GMP_RNDU);
	if (mpfr_sgn (diff) < 0)
		throw "Trying to take the square root of a negative number.";
	mpfr_sqrt (result, diff, which ? GMP_RNDD : GMP_RNDU);
	if (!which)
		mpfr_setsign (result, result, 1, GMP_RNDD);
	return;
} /* quadr_pre_lower */

/// Computes an upper bound on the preimage of a given point
/// under the quadratic map fa(x)=a-x^2, that is, computes
/// x = +/- square root of (a - y).
/// The number "y" is the argument, the number "a" is the parameter,
/// the integer indicates which part of the domain to use for the preimage:
/// 0 means to the left from the critical point, 1 -- to the right.
/// Throws an exception in case this is not possible.
inline void quadr_pre_upper (auto_mpfr &result,
	const auto_mpfr &arg, const auto_mpfr &param, int which)
{
	auto_mpfr diff;
	mpfr_sub (diff, param, arg, which ? GMP_RNDU : GMP_RNDD);
	if (mpfr_sgn (diff) < 0)
		throw "Trying to take the square root of a negative number.";
	mpfr_sqrt (result, diff, which ? GMP_RNDU : GMP_RNDD);
	if (!which)
		mpfr_setsign (result, result, 1, GMP_RNDD);
	return;
} /* quadr_pre_upper */

/// Simplified multiplication of intervals that works if both intervals
/// don't contain zero.
inline void minus2xy (auto_mpfr &leftResult, auto_mpfr &rightResult,
	const auto_mpfr &left1, const auto_mpfr &right1,
	const auto_mpfr &left2, const auto_mpfr &right2)
{
	// get the signs of the intervals
	int left1sgn = mpfr_sgn (left1);
	int left2sgn = mpfr_sgn (left2);
	int right1sgn = mpfr_sgn (right1);
	int right2sgn = mpfr_sgn (right2);

	// make sure the intervals don't contain zero
	if (left1sgn != right1sgn)
		throw "The 1st interval contains 0 in the -2xy computation!";
	if (left2sgn != right2sgn)
		throw "The 2nd interval contains 0 in the -2xy computation!";

	// compute the product without the "-2" coefficient yet;
	// note the opposite rounding direction
	if ((left1sgn > 0) && (left2sgn > 0))
	{
		mpfr_mul (leftResult, right2, right1, GMP_RNDU);
		mpfr_mul (rightResult, left2, left1, GMP_RNDD);
	}
	else if ((right1sgn < 0) && (right2sgn < 0))
	{
		mpfr_mul (leftResult, left2, left1, GMP_RNDU);
		mpfr_mul (rightResult, right2, right1, GMP_RNDD);
	}
	else if ((left1sgn > 0) && (right2sgn < 0))
	{
		mpfr_mul (leftResult, right2, left1, GMP_RNDU);
		mpfr_mul (rightResult, left2, right1, GMP_RNDD);
	}
	else if ((right1sgn < 0) && (left2sgn > 0))
	{
		mpfr_mul (leftResult, left2, right1, GMP_RNDU);
		mpfr_mul (rightResult, right2, left1, GMP_RNDD);
	}

	// multiply the resulting interval bounds by -2
	mpfr_neg (leftResult, leftResult, GMP_RNDD);
	mpfr_mul_ui (leftResult, leftResult, 2, GMP_RNDD);
	mpfr_neg (rightResult, rightResult, GMP_RNDU);
	mpfr_mul_ui (rightResult, rightResult, 2, GMP_RNDU);

	return;
} /* minus2xy */


// --------------------------------------------------
// ------------- SPLITTING THE INTERVAL -------------
// --------------------------------------------------

/// Computes a lower bound and an upper bound on C_n(a).
/// Assume zero is not in the interval that bounds all C_i(a) on the way,
/// possibly except for the final one.
/// TODO: Possible improvement: Compute one bound only (would save 50% time).
int compute_CN (auto_mpfr &cNlow, auto_mpfr &cNupp,
	const auto_mpfr &parameter, int N)
{
	// note: C_0(a) = a
	mpfr_set (cNlow, parameter, GMP_RNDD);
	mpfr_set (cNupp, parameter, GMP_RNDU);

	// prepare a temporary variable (to be used in the computations)
	auto_mpfr temp;

	// iterate the point by the function f
	for (int n = 0; n < N; ++ n)
	{
		// which side of zero is the point at?
		int sgn = mpfr_sgn (cNlow);

		// the quadratic map is increasing to the left of zero
		if (sgn <= 0)
		{
			quadr_lower (cNlow, cNlow, parameter);
			quadr_upper (cNupp, cNupp, parameter);
		}
		// the quadratic map is decreasing to the right of zero
		else
		{
			quadr_lower (temp, cNupp, parameter);
			quadr_upper (cNupp, cNlow, parameter);
			mpfr_set (cNlow, temp, GMP_RNDD);
		}

		// just in case: check if the interval is correct
		if (mpfr_cmp (cNlow, cNupp) > 0)
		{
			throw "WRONG interval detected in compute_CN.\n";
		}
	}

	return 0;
} /* compute_CN */

/// Finds a possibly minimal (or maximal) value of the parameter "a"
/// such that "C_n(a)" is below (or above) a certain border value,
/// depending on whether C_n is increasing or decreasing.
/// The returned value belongs to the given interval omega.
/// The border value in the image of C_n is given, and should be contained
/// in the inner bound for the interval omega.
/// The information on whether C_n is increasing or decreasing is provided.
/// The new parameter is found below or above the border value, as requested.
/// TODO: Possible improvement for the future: Given the images
/// of leftOmega and rightOmega, an initial guess for the parameter
/// can be made using the proportion of where the border is located.
/// Returns 0.
int find_best_parameter (auto_mpfr &parameter,
	const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
	const auto_mpfr &border, int N,
	bool increasing, bool below, int bisectionSteps)
{
	bool decreasing = !increasing;
	bool above = !below;

	// prepare the dynamic endpoints for the bisection method
	auto_mpfr left;
	mpfr_set (left, leftOmega, GMP_RNDD);
	auto_mpfr right;
	mpfr_set (right, rightOmega, GMP_RNDU);

	// prepare variables for the computations
	auto_mpfr middle;
	auto_mpfr cNlow, cNupp;

	// run the bisection method to tighten down the interval
	for (int step = bisectionSteps - 1; step >= 0; -- step)
	{
		// compute a middle point between the two endpoints
		mpfr_add (middle, left, right, GMP_RNDN);
		mpfr_div_ui (middle, middle, 2, GMP_RNDN);

		// compute the image of the middle point by C_n
		compute_CN (cNlow, cNupp, middle, N);

		// decide how to tighten the interval
		if (increasing && below)
		{
			// all the time: upp(left) <= border
			// possibly: upp(right) > border
			// and eventually return left
			if (mpfr_cmp (cNupp, border) <= 0)
				mpfr_set (left, middle, GMP_RNDD);
			else
				mpfr_set (right, middle, GMP_RNDD);
			if (step == 0)
				mpfr_set (parameter, left, GMP_RNDD);
		}
		else if (increasing && above)
		{
			// all the time: low(right) >= border
			// possibly: low(left) < border
			// and eventually return right
			if (mpfr_cmp (cNlow, border) >= 0)
				mpfr_set (right, middle, GMP_RNDU);
			else
				mpfr_set (left, middle, GMP_RNDU);
			if (step == 0)
				mpfr_set (parameter, right, GMP_RNDU);
		}
		else if (decreasing && below)
		{
			// all the time: upp(right) <= border
			// possibly: upp(left) > border
			// and eventually return right
			if (mpfr_cmp (cNupp, border) <= 0)
				mpfr_set (right, middle, GMP_RNDU);
			else
				mpfr_set (left, middle, GMP_RNDU);
			if (step == 0)
				mpfr_set (parameter, right, GMP_RNDU);
		}
		else if (decreasing && above)
		{
			// all the time: low(left) >= border
			// possibly: low(right) < border
			// and eventually return left
			if (mpfr_cmp (cNlow, border) >= 0)
				mpfr_set (left, middle, GMP_RNDD);
			else
				mpfr_set (right, middle, GMP_RNDD);
			if (step == 0)
				mpfr_set (parameter, left, GMP_RNDD);
		}
	}

	return 0;
} /* find_best_parameter */

/// Checks if omega_N hit Delta. If it doesn't then returns 0,
/// and the iterations of the interval can be continued.
/// Otherwise returns 1 and the iterations of the interval
/// should be stopped.
/// If necessary, splits the interval omega, adds new intervals to the queue
/// for further iterations (managed by the Controller), before returning 1.
/// Note: The fact whether C_n is increasing or not can be determined
/// by checking the sign of C'_n, and must be passed to this function.
int check_omega_hit_delta
	(const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
	const auto_mpfr &delta, const auto_mpfr &minusDelta,
	Controller &ctrl, int N,
//	const auto_mpfr &leftNlow, const auto_mpfr &leftNupp,
//	const auto_mpfr &rightNlow, const auto_mpfr &rightNupp,
	const auto_mpfr &leftNinner, const auto_mpfr &leftNouter,
	const auto_mpfr &rightNinner, const auto_mpfr &rightNouter,
//	const auto_mpfr &leftNminus1, const auto_mpfr &rightNminus1,
	const auto_mpfr &lowC, const auto_mpfr &uppC,
	const auto_mpfr &lowF, const auto_mpfr &uppF,
	const auto_mpfr &lowD, const auto_mpfr &uppD,
	bool increasing)
{
	// if the interval doesn't intersect Delta then return 0
	if (mpfr_cmp (rightNouter, minusDelta) <= 0)
		return 0;
	if (mpfr_cmp (leftNouter, delta) >= 0)
		return 0;

	// notify the controller of the fact that omega_n hit Delta
	int satisfied = ctrl. omega_hit_delta (N, leftOmega, rightOmega,
		delta, minusDelta,
		leftNinner, leftNouter, rightNinner, rightNouter,
		lowC, uppC, lowF, uppF, lowD, uppD, increasing);

	// if the controller is satisfied with this interval
	// then no more processing is necessary; return 1 right now
	if (satisfied)
		return 1;

	// prepare two new intervals for continuing the computations:
	// [leftOmega, leftMiddle] and [rightMiddle, rightOmega]
	auto_mpfr leftMiddle;
	mpfr_set (leftMiddle, leftOmega, GMP_RNDD);
	auto_mpfr rightMiddle;
	mpfr_set (rightMiddle, rightOmega, GMP_RNDU);

	// if there is room to the left of Delta for continuing iterations
	if (mpfr_cmp (leftNinner, minusDelta) < 0)
	{
		find_best_parameter (increasing ? leftMiddle : rightMiddle,
			leftOmega, rightOmega, minusDelta, N,
			increasing, true, ctrl. bisectionSteps ());
	}

	// if there is room to the right of Delta for continuing iterations
	if (mpfr_cmp (rightNinner, delta) > 0)
	{
		find_best_parameter (increasing ? rightMiddle : leftMiddle,
			leftOmega, rightOmega, delta, N,
			increasing, false, ctrl. bisectionSteps ());
	}

	// add the two new intervals to the queue for further computations
	// (only if they are nontrivial and large enough)
	ctrl. add_interval_parts (N, leftOmega, rightOmega,
		leftMiddle, rightMiddle);

	// make a note of the exclusion of the parameters
	ctrl. interval_excluded (N, leftOmega, rightOmega,
		leftMiddle, rightMiddle);

	return 1;
} /* check_omega_hit_delta */


// --------------------------------------------------
// ------------- ITERATIONS OF THE MAP --------------
// --------------------------------------------------

/// Initializes the iterations for the given interval omega.
int iterations_init
	(const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
	auto_mpfr &leftNlow, auto_mpfr &leftNupp,
	auto_mpfr &rightNlow, auto_mpfr &rightNupp,
	auto_mpfr &leftNinner, auto_mpfr &leftNouter,
	auto_mpfr &rightNinner, auto_mpfr &rightNouter,
	auto_mpfr &leftNminus1, auto_mpfr &rightNminus1,
	auto_mpfr &lowC, auto_mpfr &uppC,
	auto_mpfr &lowF, auto_mpfr &uppF,
	auto_mpfr &lowD, auto_mpfr &uppD)
{
	// choose which are the actual bounds of the interval \omega_N:
	// the left bound of this interval (inner and outer bounds),
	// and the right bound of this interval (also both bounds);
	// omega_0 (in the phase space) equals omega (in the parameter space)
	mpfr_set (leftNouter, leftOmega, GMP_RNDD);
	mpfr_set (leftNinner, leftOmega, GMP_RNDU);
	mpfr_set (rightNouter, rightOmega, GMP_RNDU);
	mpfr_set (rightNinner, rightOmega, GMP_RNDD);

	// initialize the outer bounds on \omega_{N-1}
	mpfr_set (leftNminus1, leftOmega, GMP_RNDD);
	mpfr_set (rightNminus1, rightOmega, GMP_RNDU);

	// initialize the low/upp bounds on the iterates of the left
	// and the right endpoints of the initial interval omega
	mpfr_set (leftNlow, leftOmega, GMP_RNDD);
	mpfr_set (leftNupp, leftOmega, GMP_RNDU);
	mpfr_set (rightNlow, rightOmega, GMP_RNDD);
	mpfr_set (rightNupp, rightOmega, GMP_RNDU);

	// create high-precision numbers for the two bounds of C'_n
	mpfr_set_ui (lowC, 1, GMP_RNDN);
	mpfr_set_ui (uppC, 1, GMP_RNDN);

	// create high-precision numbers for the two bounds of (f^n)'
	mpfr_set_ui (lowF, 1, GMP_RNDN);
	mpfr_set_ui (uppF, 1, GMP_RNDN);

	// initialize the two bounds of the quotient D
	mpfr_set_ui (lowD, 1, GMP_RNDN);
	mpfr_set_ui (uppD, 1, GMP_RNDN);

	return 0;
} /* iterations_init */

/// Makes the next iteration. On the input, obtains data for the current N.
/// Checks if the iteration can be made safely and the result is feasible.
/// Returns -1 if the iterations must be abandoned,
/// e.g., because there is a mismatch in the sign of C'_{N+1},
/// or the inner part of the interval got empty, or the interval hit Delta.
/// After having completed the iteration,
/// updates the data so that it is valid for N+1, and returns 0.
/// There might be other problems that are checked by the controller, such as:
/// - omega_N intersects Delta (confirmed!),
/// - omega_N might intersect Delta, but this is not confirmed,
/// - the inner bound on the interval shrunk to the empty set.
/// The data is: the bounds of the original interval omega,
/// delta and -delta, verbosity level, current N (needs to be increased),
/// lower and upper bounds on the Nth iterate of the endpoints of omega,
/// pointers indicating which are outer and inner bounds of omegaN
/// (that is, left bound of omega N, outer bound is the most left endpoint),
/// outer bounds on the previous iteration of the interval (omega N-1),
/// lower and upper bounds on the values C, F, and D.
int next_iteration
	(const auto_mpfr &leftOmega, const auto_mpfr &rightOmega,
	const auto_mpfr &delta, const auto_mpfr &minusDelta,
	Controller &ctrl, int N,
	auto_mpfr &leftNlow, auto_mpfr &leftNupp,
	auto_mpfr &rightNlow, auto_mpfr &rightNupp,
	auto_mpfr &leftNinner, auto_mpfr &leftNouter,
	auto_mpfr &rightNinner, auto_mpfr &rightNouter,
	auto_mpfr &leftNminus1, auto_mpfr &rightNminus1,
	auto_mpfr &lowC, auto_mpfr &uppC,
	auto_mpfr &lowF, auto_mpfr &uppF,
	auto_mpfr &lowD, auto_mpfr &uppD)
{
	// prepare temporary variables (to be re-used without initialization)
	auto_mpfr temp;
	auto_mpfr square;

	// notify the controller about the bounds on the derivative
	// and quit the iterations if ordered to
	if (ctrl. check_derivative (N, leftOmega, rightOmega,
		leftNouter, rightNouter, lowF, uppF) != 0)
	{
		return -1;
	}

	// compute the bounds on C'_{n+1}
	auto_mpfr lowCnext;
	auto_mpfr uppCnext;
	try
	{
		minus2xy (lowCnext, uppCnext,
			lowC, uppC, leftNouter, rightNouter);
		mpfr_add_si (lowCnext, lowCnext, 1, GMP_RNDD);
		mpfr_add_si (uppCnext, uppCnext, 1, GMP_RNDU);
	}
	catch(...)
	{
		ctrl. minus2xyError (N, "C", lowC, uppC,
			leftNouter, rightNouter);
		throw;
	}

	// notify the controller about the computed interval
	ctrl. interval_computed (N + 1, "C", lowCnext, uppCnext);

	// check the sign of both endpoints of the bounds on C'_{n+1}
	int lowCsgn = mpfr_sgn (lowCnext);
	int uppCsgn = mpfr_sgn (uppCnext);
	if (!((lowCsgn > 0) && (uppCsgn > 0)) &&
		!((lowCsgn < 0) && (uppCsgn < 0)))
	{
		ctrl. signMismatch (N + 1, Reason::problemC,
			leftOmega, rightOmega);
		return -1;
	}

	// -------------------------------------------------------
	// update all the variables now to make the next iteration
	// -------------------------------------------------------

	// compute forward images of the endpoints
	int sgnN = mpfr_sgn (leftNupp);
	if (sgnN < 0) // if the function f is increasing
	{
		quadr_lower (leftNlow, leftNlow, leftOmega);
		quadr_upper (leftNupp, leftNupp, leftOmega);
		quadr_lower (rightNlow, rightNlow, rightOmega);
		quadr_upper (rightNupp, rightNupp, rightOmega);
	}
	else // if the function f is decreasing
	{
		quadr_lower (temp, leftNupp, leftOmega);
		quadr_upper (leftNupp, leftNlow, leftOmega);
		mpfr_swap (leftNlow, temp);
		quadr_lower (temp, rightNupp, rightOmega);
		quadr_upper (rightNupp, rightNlow, rightOmega);
		mpfr_swap (rightNlow, temp);
	}

	// notify the controller about the computed intervals
	ctrl. interval_computed (N + 1, "leftOmega", leftNlow, leftNupp);
	ctrl. interval_computed (N + 1, "rightOmega", rightNlow, rightNupp);

	// store the outer bounds on omega_n for the next iterate
	mpfr_set (leftNminus1, leftNouter, GMP_RNDN);
	mpfr_set (rightNminus1, rightNouter, GMP_RNDN);

	// remember which bounds are outer and inner at n+1
	if (lowCsgn > 0) // if the function c_n is increasing
	{
		mpfr_set (leftNouter, leftNlow, GMP_RNDD);
		mpfr_set (leftNinner, leftNupp, GMP_RNDU);
		mpfr_set (rightNouter, rightNupp, GMP_RNDU);
		mpfr_set (rightNinner, rightNlow, GMP_RNDD);
	}
	else // if the function c_n is decreasing
	{
		mpfr_set (leftNouter, rightNlow, GMP_RNDD);
		mpfr_set (leftNinner, rightNupp, GMP_RNDU);
		mpfr_set (rightNouter, leftNupp, GMP_RNDU);
		mpfr_set (rightNinner, leftNlow, GMP_RNDD);
	}

	// notify the controller about the computed outer bounds
	ctrl. interval_computed (N + 1, "outerOmega",
		leftNouter, rightNouter);

	// check if the inner bound of the interval got empty
	if (mpfr_cmp (leftNinner, rightNinner) >= 0)
	{
		ctrl. innerBoundEmpty (N + 1, leftOmega, rightOmega);
		return -1;
	}

	// compute high-precision bounds on (f^{n+1})'
	auto_mpfr lowFnext;
	auto_mpfr uppFnext;
	try
	{
		minus2xy (lowFnext, uppFnext,
			lowF, uppF, leftNminus1, rightNminus1);
	}
	catch(...)
	{
		ctrl. minus2xyError (N, "F", lowF, uppF,
			leftNminus1, rightNminus1);
		throw;
	}

	// notify the controller about the computed interval
	ctrl. interval_computed (N + 1, "F", lowFnext, uppFnext);

	// check the sign of both endpoints of the bounds on (f^{n+1})'
	int lowFsgn = mpfr_sgn (lowFnext);
	int uppFsgn = mpfr_sgn (uppFnext);
	if (!((lowFsgn > 0) && (uppFsgn > 0)) &&
		!((lowFsgn < 0) && (uppFsgn < 0)))
	{
		ctrl. signMismatch (N + 1, Reason::problemF,
			leftOmega, rightOmega);
		return -1;
	}

	// compute both bounds on D_{n+1} := D_n + 1/(f^n+1)'
	auto_mpfr lowDnext;
	mpfr_set (lowDnext, lowD, GMP_RNDD);
	mpfr_ui_div (temp, 1, uppF, GMP_RNDD);
	mpfr_add (lowDnext, lowD, temp, GMP_RNDD);
	auto_mpfr uppDnext;
	mpfr_set (uppDnext, uppD, GMP_RNDU);
	mpfr_ui_div (temp, 1, lowF, GMP_RNDU);
	mpfr_add (uppDnext, uppD, temp, GMP_RNDU);

	// notify the controller about the computed interval
	ctrl. interval_computed (N + 1, "D", lowDnext, uppDnext);

	// store the new computed data to complete this iteration
	mpfr_swap (lowC, lowCnext);
	mpfr_swap (uppC, uppCnext);
	mpfr_swap (lowF, lowFnext);
	mpfr_swap (uppF, uppFnext);
	mpfr_swap (lowD, lowDnext);
	mpfr_swap (uppD, uppDnext);

	// notify the controller that this iteration has almost been completed
	ctrl. iteration_completed (N + 1, leftNinner, leftNouter,
		rightNinner, rightNouter);

	// ckeck if the new interval intersects Delta,
	// and split the interval if necessary
	int hit_delta = check_omega_hit_delta (leftOmega, rightOmega,
		delta, minusDelta, ctrl, N+1,
	//	leftNlow, leftNupp, rightNlow, rightNupp,
		leftNinner, leftNouter, rightNinner, rightNouter,
	//	leftNminus1, rightNminus1,
		lowC, uppC, lowF, uppF, lowD, uppD,
		mpfr_sgn (lowC) >= 0);
	if (hit_delta)
		return -1;

	return 0;
} /* next_iteration */


// --------------------------------------------------
// ---------------- MAIN COMPUTATION ----------------
// --------------------------------------------------

/// Runs the main computation for the quadratic map.
void quadratic_map_iterations (const char *paramLeftStr,
	const char *paramRightStr, int uniform,
	const char *deltaStr, int precision, int bisectionSteps,
	int maxIter, int maxIntervals, int maxQueueSize,
	int maxMessages, int timeLimit,
	int targetN, int requiredN, const char *requiredWidthStr,
	int showWorstIntervals,
	const char *fractionStr, const char *minWidthStr,
	int contractions, int outDigits, int verbosity)
{
	using chomp::homology::sout;

	// set the default precision to be used for all the computations
	auto_mpfr::setDefaultPrecision (precision);

	// prepare the endpoints of the full interval Omega
	auto_mpfr leftOmegaFull;
	if (mpfr_set_str (leftOmegaFull, paramLeftStr, 10, GMP_RNDN) < 0)
		throw "The left parameter bound is not a valid number.";
	auto_mpfr rightOmegaFull;
	if (mpfr_set_str (rightOmegaFull, paramRightStr, 10, GMP_RNDN) < 0)
		throw "The right parameter bound is not a valid number.";
	if (mpfr_cmp (leftOmegaFull, rightOmegaFull) > 0)
		throw "The left parameter is greater than the right one.";

	// define the minimal width of intervals allowed in the queue
	auto_mpfr minWidth;
	auto_mpfr minWidthFraction;
	if (mpfr_set_str (minWidthFraction, minWidthStr, 10, GMP_RNDN) < 0)
		throw "Wrong minimal width requested.";
	if (mpfr_sgn (minWidthFraction) < 0)
		throw "Negative minimal width requested.";
	else if (mpfr_sgn (minWidthFraction) > 0)
	{
		mpfr_sub (minWidth, rightOmegaFull, leftOmegaFull, GMP_RNDD);
		mpfr_mul (minWidth, minWidth, minWidthFraction, GMP_RNDD);
	}
	else
	{
		mpfr_set_ui (minWidth, 0, GMP_RNDD);
	}

	// determine the radius of the critical neighborhood
	auto_mpfr delta;
	if (mpfr_set_str (delta, deltaStr, 10, GMP_RNDU))
		throw "The number delta is not a valid number.";
	if (mpfr_sgn (delta) <= 0)
		throw "The number delta seems to be non-positive.";

	// determine the fraction requested
	auto_mpfr fraction;
	if (mpfr_set_str (fraction, fractionStr, 10, GMP_RNDU))
		throw "The fraction requested is not a valid number.";
	if (mpfr_sgn (fraction) < 0)
		throw "The requested fraction seems to be negative.";

	// prepare (a lower bound on) the number -\delta
	auto_mpfr minusDelta;
	mpfr_neg (minusDelta, delta, GMP_RNDD);

	// prepare the high-precision number for the required width of omega_n
	auto_mpfr requiredWidth;
	if (mpfr_set_str (requiredWidth, requiredWidthStr, 10, GMP_RNDU))
		throw "The required width requested is not a valid number.";
	if (mpfr_sgn (requiredWidth) < 0)
		throw "The requested required width seems to be negative.";

	// prepare a controller object for keeping track of
	// and controlling the computations
	Controller ctrl (leftOmegaFull, rightOmegaFull,
		maxIter, maxIntervals, maxQueueSize, maxMessages,
		timeLimit, targetN, requiredN, requiredWidth,
		showWorstIntervals, fraction, minWidth, contractions,
		bisectionSteps, outDigits, verbosity);

	// add the initial interval or intervals to the controller
	if (uniform <= 1)
	{
		ctrl. add (leftOmegaFull, rightOmegaFull, 0, Reason::init);
	}
	else
	{
		// compute the size of each part of the full interval
		auto_mpfr partSize;
		mpfr_sub (partSize, rightOmegaFull, leftOmegaFull, GMP_RNDN);
		mpfr_div_ui (partSize, partSize, uniform, GMP_RNDN);

		// prepare variables for the endpoints of small intervals
		auto_mpfr x, y;

		// set the leftmost endpoint
		mpfr_set (x, leftOmegaFull, GMP_RNDN);

		// use the endpoints interchangeably
		// to add all the small intervals to the queue
		bool which (true);
		for (int i = 1; i <= uniform; ++ i)
		{
			if (i == uniform)
			{
				mpfr_set (which ? y : x, rightOmegaFull,
					GMP_RNDN);
			}
			else
			{
				mpfr_mul_ui (which ? y : x, partSize, i,
					GMP_RNDN);
				mpfr_add (which ? y : x, which ? y : x,
					leftOmegaFull, GMP_RNDN);
			}
			ctrl. add (which ? x : y, which ? y : x, 0,
				Reason::init);
			which = !which;
		}
	}

	// prepare variables to be used for iterations
	auto_mpfr leftOmega, rightOmega;
	auto_mpfr leftNlow, leftNupp;
	auto_mpfr rightNlow, rightNupp;
	auto_mpfr leftNouter, leftNinner;
	auto_mpfr rightNouter, rightNinner;
	auto_mpfr leftNminus1, rightNminus1;
	auto_mpfr lowC, uppC;
	auto_mpfr lowF, uppF;
	auto_mpfr lowD, uppD;

	sout << "Running the computations (verbosity level: " <<
		verbosity << " out of 100).\n\n";

	// process all the intervals in the queue managed by the controller
	while (ctrl. size ())
	{
		// before taking the next interval, check if the limits
		// on time, the number of intervals, etc. have been reached
		if (ctrl. check_iterating (-1, leftOmega, rightOmega) != 0)
			break;

		// get an interval for processing from the queue
		ctrl. get (leftOmega, rightOmega);

		// initialize the iterations for the entire interval omega
		iterations_init (leftOmega, rightOmega,
			leftNlow, leftNupp, rightNlow, rightNupp,
			leftNinner, leftNouter, rightNinner, rightNouter,
			leftNminus1, rightNminus1,
			lowC, uppC, lowF, uppF, lowD, uppD);

		// compute the iterations until something happens
		for (int N = 0;; ++ N)
		{
			// check if the number of iterations
			// or the time limit has been exceeded
			if (ctrl. check_iterating (N,
				leftOmega, rightOmega) != 0)
			{
				break;
			}

			// compute omega_{N+1} and the related quantities
			int result = next_iteration (leftOmega, rightOmega,
				delta, minusDelta, ctrl, N,
				leftNlow, leftNupp, rightNlow, rightNupp,
				leftNinner, leftNouter,
				rightNinner, rightNouter,
				leftNminus1, rightNminus1,
				lowC, uppC, lowF, uppF, lowD, uppD);

			// exit the loop if anything went wrong
			if (result != 0)
				break;
		}
	}

	// clear the queue and show a summary of what was done
	ctrl. finalize ();

	return;
} /* quadratic_map_iterations */


// --------------------------------------------------
// ---------------------- MAIN ----------------------
// --------------------------------------------------

/// The main function of the program.
/// Returns: 0 = Ok, -1 = Error, 1 = Help displayed, 2 = Wrong arguments.
int main (int argc, char *argv [])
{
	using namespace chomp::homology;

	// prepare user-configurable data
	char *paramLeftStr = 0;
	char *paramRightStr = 0;
	int uniform = 1;
	char *deltaStr = const_cast<char *> ("0.001");
	int precision = 300;
	int bisectionSteps = 40;
	int maxIterations = 200;
	char *fraction = const_cast<char *> ("0");
	char *minWidth = const_cast<char *> ("1e-10");
	int contractions = 0;
	int maxIntervals = 0;
	int maxQueueSize = 0;
	int maxMessages = 0;
	int timeLimit = 10;
	int targetN = 0;
	int requiredN = 0;
	char *requiredWidth = const_cast<char *> ("0");
	int outDigits = 8;
	int verbosity = 0;
	int showWorstIntervals = 0;

	// interprete the command-line arguments
	arguments a;
	arg (a, "l", paramLeftStr);
	arg (a, "r", paramRightStr);
	arg (a, "u", uniform);
	arg (a, "d", deltaStr);
	arg (a, "p", precision);
	arg (a, "b", bisectionSteps);
	arg (a, "n", maxIterations);
	arg (a, "f", fraction);
	arg (a, "w", minWidth);
	arg (a, "c", contractions);
	arg (a, "i", maxIntervals);
	arg (a, "q", maxQueueSize);
	arg (a, "m", maxMessages);
	arg (a, "t", timeLimit);
	arg (a, "N", targetN);
	arg (a, "rN", requiredN);
	arg (a, "rW", requiredWidth);
	arg (a, "o", outDigits);
	arg (a, "v", verbosity);
	arg (a, "-show-worst-intervals", showWorstIntervals);
	argswitch (a, "-show-exclusion-data", Verbosity::exclusionData, 0);
	argswitch (a, "-show-hit-data", Verbosity::hitData, 0);
	argswitch (a, "-show-good-data", Verbosity::goodData, 0);
	argswitch (a, "-show-iterated-data", Verbosity::iteratedData, 0);
	argswitch (a, "-show-derivative-data", Verbosity::derivativeData, 0);
	argswitch (a, "-problems-first", problemsFirst, true);
	arghelp (a);

	argstreamprepare (a);
	int argresult = a. analyze (argc, argv);
	argstreamset ();

	// show the program's main title
	if (argresult >= 0)
		sout << title << '\n';

	// if something was incorrect, show an additional message and exit
	if (argresult < 0)
	{
		sout << "Call with '--help' for help.\n";
		return 2;
	}

	// make sure that the parameters were defined
	if (!paramLeftStr || !paramRightStr)
	{
		if (argc > 1)
		{
			sout << "WARNING: Both bounds of the " <<
				"paramter interval must be defined.\n";
		}
		argresult = 1;
	}

	// make sure the number of intervals in the initial subdivision is OK
	if (uniform < 1)
	{
		sout << "WARNING: The number of intervals "
			"in the uniform subdivision must be at least 1.\n";
		argresult = 1;
	}

	// make sure the binary precision of the computations is correct
	if ((precision < 100) || (precision > 100000))
	{
		sout << "WARNING: The binary precision must be " <<
			"between 100 and 100000.\n";
		argresult = 1;
	}

	// make sure the number of bisection steps is reasonable
	if ((bisectionSteps < 4) || (bisectionSteps > 100))
	{
		sout << "WARNING: The number of bisection steps must be " <<
			"between 4 and 100; at least 20 recommended.\n";
		argresult = 1;
	}

	// make sure the max number of iterations allowed is good
	if ((maxIterations < 0) || (maxIterations > 100000))
	{
		sout << "WARNING: The max number of iterations must be " <<
			"between 0 and 100000.\n";
		argresult = 1;
	}

	// make sure the number of output digits is within the allowed limits
	if ((outDigits < 1) || (outDigits > 1000))
	{
		sout << "WARNING: The number of output digits must be " <<
			"between 1 and 1000.\n";
		argresult = 1;
	}

	// if help requested, show help information
	if (argresult > 0)
	{
		sout << helpinfo << '\n';
		return 1;
	}

	// try running the main function and catch an error message if thrown
	try
	{
		// set an appropriate program time message
		program_time = "Aborted after:";
		program_time = 1;

		// if verbosity is at least 1 then show the parameters
		if (verbosity >= Verbosity::programArguments)
		{
			sout << "\nEffective arguments, based on "
				"the defaults and the command line:\n";
			sout << " -l " << paramLeftStr <<
				" -r " << paramRightStr <<
				" -u " << uniform <<
				" -d " << deltaStr <<
				" -p " << precision <<
				" -b " << bisectionSteps <<
				" -n " << maxIterations <<
				" -f " << fraction <<
				" -w " << minWidth <<
				" -c " << contractions <<
				" -i " << maxIntervals <<
				" -q " << maxQueueSize <<
				" -m " << maxMessages <<
				" -t " << timeLimit <<
				" -N " << targetN <<
				" -o " << outDigits <<
				" -v " << verbosity << "\n";
			sout << "Explanation of these arguments:\n";
			sout << "Part 1. Quantities that primarily determine "
				"what is being computed.\n";
			sout << " omega = [" << paramLeftStr <<
				"," << paramRightStr << "]\n";
			if (uniform != 1)
			{
				sout << " Note: This interval will be "
					"uniformly split into " << uniform <<
					" subintervals.\n";
			}
			sout << " delta = " << deltaStr << "\n";
			sout << "Part 2. Quality of numerical "
				"computations.\n";
			sout << " number of binary digits to use: " <<
				precision << "\n";
			sout << " bisection steps when splitting omega "
				"when omega_n hits Delta: " <<
				bisectionSteps << "\n";
			sout << "Part 3. Optimization features that help use "
				"the resources effectively.\n";
			sout << " max allowed number of iterations "
				"for an individual interval: " <<
				maxIterations << "\n";
			sout << " minimal fraction of the original interval "
				"required after omega_n hits Delta: " <<
				fraction << "\n";
			sout << " minimal width (relative to "
				"the initial one) required to add "
				"an interval to the queue: " <<
				minWidth << "\n";
			sout << " minimal number of decreases in F_n' "
				"required to abandon an interval: " <<
				contractions << "\n";
			sout << "Part 4. Restrictions on the amount of "
				"work to be done by the program.\n";
			sout << " max allowed number of intervals "
				"to iterate: " << maxIntervals << "\n";
			sout << " max allowed size of the queue (in order to "
				"avoid excessive memory usage): " <<
				maxQueueSize << "\n";
			sout << " max number of messages to display: " <<
				maxMessages << "\n";
			sout << " max amount of time allowed to use: " <<
				timeLimit << "\n";
			sout << " the minimal number of iterates "
				"to reach by all to stop: " <<
				targetN << "\n";
			sout << "Part 5. Escape time determination.\n";
			sout << " the required number of iterates "
				"to count as an 'escape': " <<
				requiredN << "\n";
			sout << " the required image width "
				"to count as an 'escape': " <<
				requiredWidth << "\n";
			sout << "Part 6. Amount of program's output.\n";
			sout << " number of decimal digits to display: " <<
				outDigits << "\n";
			sout << " verbosity of program's output (0-100): " <<
				verbosity << "\n";
			sout << "\n";
		}

		// run the computation with the parameters provided
		quadratic_map_iterations (paramLeftStr, paramRightStr,
			uniform, deltaStr, precision, bisectionSteps,
			maxIterations, maxIntervals, maxQueueSize,
			maxMessages, timeLimit, targetN,
			requiredN, requiredWidth, showWorstIntervals,
			fraction, minWidth, contractions,
			outDigits, verbosity);

		// set an appropriate program time message
		program_time = "Total time used:";

		// finalize
		return 0;
	}
	catch (const char *msg)
	{
		sout << "ERROR: " << msg << '\n';
		return -1;
	}
	catch (const std::exception &e)
	{
		sout << "ERROR: " << e. what () << '\n';
		return -1;
	}
	catch (...)
	{
		sout << "ABORT: An unknown error occurred.\n";
		return -1;
	}
} /* main */

